#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Synthetic functions for optimization benchmarks.

Most test functions (if not indicated otherwise) are taken from
[Bingham2013virtual]_.


References:

.. [Bingham2013virtual]
    D. Bingham, S. Surjanovic. Virtual Library of Simulation Experiments.
    https://www.sfu.ca/~ssurjano/optimization.html

.. [CoelloCoello2002constraint]
    C. A. Coello Coello and E. Mezura Montes. Constraint-handling in genetic
    algorithms through the use of dominance-based tournament selection.
    Advanced Engineering Informatics, 16(3):193–203, 2002.

.. [Gramacy2016]
    R. Gramacy, G. Gray, S. Le Digabel, H. Lee, P. Ranjan, G. Wells & S. Wild.
    Modeling an Augmented Lagrangian for Blackbox Constrained Optimization,
    Technometrics, 2016.

.. [Hedar2006derivfree]
    A.-R. Hedar and M. Fukushima. Derivative-free filter simulated annealing
    method for constrained continuous global optimization. Journal of Global
    Optimization, 35(4):521–549, 2006.

.. [Lemonge2010constrained]
    A. C. C. Lemonge, H. J. C. Barbosa, C. C. H. Borges, and F. B. dos Santos
    Silva. Constrained optimization problems in mechanical engineering design
    using a real-coded steady-state genetic algorithm. Mecánica Computacional,
    XXIX:9287–9303, 2010.

.. [Letham2019]
    B. Letham, B. Karrer, G. Ottoni, and E. Bakshy. Constrained Bayesian
    Optimization with Noisy Experiments. Bayesian Analysis, Bayesian Anal.
    14(2), 495-519, 2019.

.. [Mishra2007]
    S. K. Mishra. Minimization of Keane's Bump Function by the Repulsive
    Particle Swarm and the Differential Evolution Methods (May 1, 2007).
    Available at SSRN: https://ssrn.com/abstract=983836.

.. [Packebusch2016]
    T. Packebusch, S. Mertens. Low autocorrelation binary sequences. Journal of
    Physics A: Mathematical and Theoretical 49.16 (2016).
"""

from __future__ import annotations

import math
from abc import ABC

import torch
from botorch.exceptions.errors import InputDataError
from botorch.test_functions.base import BaseTestProblem, ConstrainedBaseTestProblem
from botorch.test_functions.utils import round_nearest
from torch import Tensor


class SyntheticTestFunction(BaseTestProblem, ABC):
    r"""Base class for synthetic test functions."""

    _optimal_value: float | None = None
    _optimizers: list[tuple[float, ...]] | None = None
    num_objectives: int = 1

    def __init__(
        self,
        noise_std: None | float | list[float] = None,
        negate: bool = False,
        bounds: list[tuple[float, float]] | None = None,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            noise_std: Standard deviation of the observation noise. If a list is
                provided, specifies separate noise standard deviations for each
                objective in a multiobjective problem.
            negate: If True, negate the function.
            bounds: Custom bounds for the function specified as (lower, upper) pairs.
            dtype: The dtype that is used for the bounds of the function.
        """
        if bounds is not None:
            self._bounds = bounds
        super().__init__(noise_std=noise_std, negate=negate, dtype=dtype)
        if self._optimizers is not None:
            if bounds is not None:
                # Ensure at least one optimizer lies within the custom bounds
                def in_bounds(
                    optimizer: tuple[float, ...], bounds: list[tuple[float, float]]
                ) -> bool:
                    for i, xopt in enumerate(optimizer):
                        lower, upper = bounds[i]
                        if xopt < lower or xopt > upper:
                            return False

                    return True

                if not any(
                    in_bounds(optimizer=optimizer, bounds=bounds)
                    for optimizer in self._optimizers
                ):
                    raise ValueError(
                        "No global optimum found within custom bounds. Please specify "
                        "bounds which include at least one point in "
                        f"`{self.__class__.__name__}._optimizers`."
                    )
            self.register_buffer(
                "optimizers", torch.tensor(self._optimizers, dtype=self.bounds.dtype)
            )

    @property
    def optimal_value(self) -> float:
        r"""The global optimum of the function.

        This can be either the minimum or maximum depending the value of
        `self.is_minimization_problem`.
        """
        if self._optimal_value is not None:
            return -self._optimal_value if self.negate else self._optimal_value
        else:
            raise NotImplementedError(
                f"Problem {self.__class__.__name__} does not specify an optimal value."
            )


class Ackley(SyntheticTestFunction):
    r"""Ackley test function.

    d-dimensional function (usually evaluated on `[-32.768, 32.768]^d`):

        f(x) = -A exp(-B sqrt(1/d sum_{i=1}^d x_i^2)) -
            exp(1/d sum_{i=1}^d cos(c x_i)) + A + exp(1)

    f has one minimizer for its global minimum at `z_1 = (0, 0, ..., 0)` with
    `f(z_1) = 0`.
    """

    _optimal_value = 0.0
    _check_grad_at_opt: bool = False

    def __init__(
        self,
        dim: int = 2,
        noise_std: float | None = None,
        negate: bool = False,
        bounds: list[tuple[float, float]] | None = None,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            dim: The (input) dimension.
            noise_std: Standard deviation of the observation noise.
            negate: If True, negate the function.
            bounds: Custom bounds for the function specified as (lower, upper) pairs.
        """
        self.dim = dim
        self.continuous_inds = list(range(dim))
        if bounds is None:
            bounds = [(-32.768, 32.768) for _ in range(self.dim)]
        self._optimizers = [tuple(0.0 for _ in range(self.dim))]
        super().__init__(noise_std=noise_std, negate=negate, bounds=bounds, dtype=dtype)
        self.a = 20
        self.b = 0.2
        self.c = 2 * math.pi

    def _evaluate_true(self, X: Tensor) -> Tensor:
        a, b, c = self.a, self.b, self.c
        part1 = -a * torch.exp(-torch.linalg.norm(X, dim=-1) * b / math.sqrt(self.dim))
        part2 = -(torch.exp(torch.mean(torch.cos(c * X), dim=-1)))
        return part1 + part2 + a + math.e


class Beale(SyntheticTestFunction):
    dim = 2
    continuous_inds = list(range(dim))
    _optimal_value = 0.0
    _bounds = [(-4.5, 4.5), (-4.5, 4.5)]
    _optimizers = [(3.0, 0.5)]

    def _evaluate_true(self, X: Tensor) -> Tensor:
        x1, x2 = X[..., 0], X[..., 1]
        part1 = (1.5 - x1 + x1 * x2).pow(2)
        part2 = (2.25 - x1 + x1 * x2.pow(2)).pow(2)
        part3 = (2.625 - x1 + x1 * x2.pow(3)).pow(2)
        return part1 + part2 + part3


class Branin(SyntheticTestFunction):
    r"""Branin test function.

    Two-dimensional function (usually evaluated on `[-5, 10] x [0, 15]`):

        B(x) = (x_2 - b x_1^2 + c x_1 - r)^2 + 10 (1-t) cos(x_1) + 10

    Here `b`, `c`, `r` and `t` are constants where `b = 5.1 / (4 * math.pi ** 2)`
    `c = 5 / math.pi`, `r = 6`, `t = 1 / (8 * math.pi)`
    B has 3 minimizers for its global minimum at `z_1 = (-pi, 12.275)`,
    `z_2 = (pi, 2.275)`, `z_3 = (9.42478, 2.475)` with `B(z_i) = 0.397887`.
    """

    dim = 2
    continuous_inds = list(range(dim))
    _bounds = [(-5.0, 10.0), (0.0, 15.0)]
    _optimal_value = 0.397887
    _optimizers = [(-math.pi, 12.275), (math.pi, 2.275), (9.42478, 2.475)]

    def _evaluate_true(self, X: Tensor) -> Tensor:
        t1 = (
            X[..., 1]
            - 5.1 / (4 * math.pi**2) * X[..., 0].pow(2)
            + 5 / math.pi * X[..., 0]
            - 6
        )
        t2 = 10 * (1 - 1 / (8 * math.pi)) * torch.cos(X[..., 0])
        return t1.pow(2) + t2 + 10


class Bukin(SyntheticTestFunction):
    dim = 2
    continuous_inds = list(range(dim))
    _bounds = [(-15.0, -5.0), (-3.0, 3.0)]
    _optimal_value = 0.0
    _optimizers = [(-10.0, 1.0)]
    _check_grad_at_opt: bool = False

    def _evaluate_true(self, X: Tensor) -> Tensor:
        part1 = 100.0 * torch.sqrt(torch.abs(X[..., 1] - 0.01 * X[..., 0].pow(2)))
        part2 = 0.01 * torch.abs(X[..., 0] + 10.0)
        return part1 + part2


class Cosine8(SyntheticTestFunction):
    r"""Cosine Mixture test function.

    8-dimensional function (usually evaluated on `[-1, 1]^8`):

        f(x) = 0.1 sum_{i=1}^8 cos(5 pi x_i) - sum_{i=1}^8 x_i^2

    f has one maximizer for its global maximum at `z_1 = (0, 0, ..., 0)` with
    `f(z_1) = 0.8`
    """

    dim = 8
    continuous_inds = list(range(dim))
    _bounds = [(-1.0, 1.0) for _ in range(8)]
    _optimal_value = 0.8
    _optimizers = [tuple(0.0 for _ in range(8))]
    _is_minimization_by_default = False

    def _evaluate_true(self, X: Tensor) -> Tensor:
        return torch.sum(0.1 * torch.cos(5 * math.pi * X) - X.pow(2), dim=-1)


class DropWave(SyntheticTestFunction):
    dim = 2
    continuous_inds = list(range(dim))
    _bounds = [(-5.12, 5.12), (-5.12, 5.12)]
    _optimal_value = -1.0
    _optimizers = [(0.0, 0.0)]
    _check_grad_at_opt = False

    def _evaluate_true(self, X: Tensor) -> Tensor:
        norm = torch.linalg.norm(X, dim=-1)
        part1 = 1.0 + torch.cos(12.0 * norm)
        part2 = 0.5 * norm.pow(2) + 2.0
        return -part1 / part2


class DixonPrice(SyntheticTestFunction):
    _optimal_value = 0.0

    def __init__(
        self,
        dim=2,
        noise_std: float | None = None,
        negate: bool = False,
        bounds: list[tuple[float, float]] | None = None,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            dim: The (input) dimension.
            noise_std: Standard deviation of the observation noise.
            negate: If True, negate the function.
            dtype: The dtype that is used for the bounds of the function.
        """
        self.dim = dim
        self.continuous_inds = list(range(dim))
        if bounds is None:
            bounds = [(-10.0, 10.0) for _ in range(self.dim)]
        self._optimizers = [
            tuple(
                math.pow(2.0, -(1.0 - 2.0 ** (-(i - 1))))
                for i in range(1, self.dim + 1)
            )
        ]
        super().__init__(noise_std=noise_std, negate=negate, bounds=bounds, dtype=dtype)

    def _evaluate_true(self, X: Tensor) -> Tensor:
        d = self.dim
        part1 = (X[..., 0] - 1).pow(2)
        i = X.new(range(2, d + 1))
        part2 = torch.sum(i * (2.0 * X[..., 1:].pow(2) - X[..., :-1]).pow(2), dim=-1)
        return part1 + part2


class EggHolder(SyntheticTestFunction):
    r"""Eggholder test function.

    Two-dimensional function (usually evaluated on `[-512, 512]^2`):

        E(x) = (x_2 + 47) sin(R1(x)) - x_1 * sin(R2(x))

    where `R1(x) = sqrt(|x_2 + x_1 / 2 + 47|)`, `R2(x) = sqrt|x_1 - (x_2 + 47)|)`.
    """

    dim = 2
    continuous_inds = list(range(dim))
    _bounds = [(-512.0, 512.0), (-512.0, 512.0)]
    _optimal_value = -959.6407
    _optimizers = [(512.0, 404.2319)]
    _check_grad_at_opt: bool = False

    def _evaluate_true(self, X: Tensor) -> Tensor:
        x1, x2 = X[..., 0], X[..., 1]
        part1 = -(x2 + 47.0) * torch.sin(torch.sqrt(torch.abs(x2 + x1 / 2.0 + 47.0)))
        part2 = -x1 * torch.sin(torch.sqrt(torch.abs(x1 - (x2 + 47.0))))
        return part1 + part2


class Griewank(SyntheticTestFunction):
    r"""Griewank synthetic test function.

    The Griewank function is defined for any `d`, is typically evaluated on
    `[-600, 600]^d`, and given by:

        G(x) = sum_{i=1}^d x_i**2 / 4000 - prod_{i=1}^d cos(x_i / sqrt(i)) + 1

    G has many widespread local minima, which are regularly distributed.
    The global minimum is at `z = (0, ..., 0)` with `G(z) = 0`.
    """

    _optimal_value = 0.0

    def __init__(
        self,
        dim=2,
        noise_std: float | None = None,
        negate: bool = False,
        bounds: list[tuple[float, float]] | None = None,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            dim: The (input) dimension.
            noise_std: Standard deviation of the observation noise.
            negate: If True, negate the function.
            bounds: Custom bounds for the function specified as (lower, upper) pairs.
            dtype: The dtype that is used for the bounds of the function.
        """
        self.dim = dim
        self.continuous_inds = list(range(dim))
        if bounds is None:
            bounds = [(-600.0, 600.0) for _ in range(self.dim)]
        self._optimizers = [tuple(0.0 for _ in range(self.dim))]
        super().__init__(noise_std=noise_std, negate=negate, bounds=bounds, dtype=dtype)

    def _evaluate_true(self, X: Tensor) -> Tensor:
        part1 = torch.sum(X.pow(2) / 4000.0, dim=-1)
        d = X.shape[-1]
        part2 = -(torch.prod(torch.cos(X / torch.sqrt(X.new(range(1, d + 1)))), dim=-1))
        return part1 + part2 + 1.0


class Hartmann(SyntheticTestFunction):
    r"""Hartmann synthetic test function.

    Most commonly used is the six-dimensional version (typically evaluated on
    `[0, 1]^6`):

        H(x) = - sum_{i=1}^4 ALPHA_i exp( - sum_{j=1}^6 A_ij (x_j - P_ij)**2 )

    H has a 6 local minima and a global minimum at

        z = (0.20169, 0.150011, 0.476874, 0.275332, 0.311652, 0.6573)

    with `H(z) = -3.32237`.
    """

    def __init__(
        self,
        dim=6,
        noise_std: float | None = None,
        negate: bool = False,
        bounds: list[tuple[float, float]] | None = None,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            dim: The (input) dimension.
            noise_std: Standard deviation of the observation noise.
            negate: If True, negate the function.
            bounds: Custom bounds for the function specified as (lower, upper) pairs.
            dtype: The dtype that is used for the bounds of the function.
        """
        if dim not in (3, 4, 6):
            raise ValueError(f"Hartmann with dim {dim} not defined")
        self.dim = dim
        self.continuous_inds = list(range(dim))
        if bounds is None:
            bounds = [(0.0, 1.0) for _ in range(self.dim)]
        # optimizers and optimal values for dim=4 not implemented
        optvals = {3: -3.86278, 6: -3.32237}
        optimizers = {
            3: [(0.114614, 0.555649, 0.852547)],
            6: [(0.20169, 0.150011, 0.476874, 0.275332, 0.311652, 0.6573)],
        }
        self._optimal_value = optvals.get(self.dim)
        self._optimizers = optimizers.get(self.dim)
        super().__init__(noise_std=noise_std, negate=negate, bounds=bounds, dtype=dtype)
        self.register_buffer("ALPHA", torch.tensor([1.0, 1.2, 3.0, 3.2]))
        if dim == 3:
            A = [[3.0, 10, 30], [0.1, 10, 35], [3.0, 10, 30], [0.1, 10, 35]]
            P = [
                [3689, 1170, 2673],
                [4699, 4387, 7470],
                [1091, 8732, 5547],
                [381, 5743, 8828.0],
            ]
        elif dim == 4:
            A = [
                [10, 3, 17, 3.5],
                [0.05, 10, 17, 0.1],
                [3, 3.5, 1.7, 10],
                [17, 8, 0.05, 10],
            ]
            P = [
                [1312, 1696, 5569, 124.0],
                [2329, 4135, 8307, 3736],
                [2348, 1451, 3522, 2883],
                [4047, 8828, 8732, 5743],
            ]
        elif dim == 6:
            A = [
                [10, 3, 17, 3.5, 1.7, 8],
                [0.05, 10, 17, 0.1, 8, 14],
                [3, 3.5, 1.7, 10, 17, 8],
                [17, 8, 0.05, 10, 0.1, 14],
            ]
            P = [
                [1312, 1696, 5569, 124, 8283, 5886],
                [2329, 4135, 8307, 3736, 1004, 9991],
                [2348, 1451, 3522, 2883, 3047, 6650],
                [4047, 8828, 8732, 5743, 1091, 381.0],
            ]
        else:  # pragma: no cover -- unreacheable code for pyre.
            raise NotImplementedError
        self.register_buffer("A", torch.tensor(A))
        self.register_buffer("P", torch.tensor(P))

    @property
    def optimizers(self) -> Tensor:
        if self.dim == 4:
            raise NotImplementedError()
        return super().optimizers

    def _evaluate_true(self, X: Tensor) -> Tensor:
        self.to(device=X.device, dtype=X.dtype)
        inner_sum = torch.sum(
            self.A * (X.unsqueeze(-2) - 0.0001 * self.P).pow(2), dim=-1
        )
        H = -(torch.sum(self.ALPHA * torch.exp(-inner_sum), dim=-1))
        if self.dim == 4:
            H = (1.1 + H) / 0.839
        return H


class HolderTable(SyntheticTestFunction):
    r"""Holder Table synthetic test function.

    Two-dimensional function (typically evaluated on `[0, 10] x [0, 10]`):

        `H(x) = - | sin(x_1) * cos(x_2) * exp(| 1 - ||x|| / pi | ) |`

    H has 4 global minima with `H(z_i) = -19.2085` at

        z_1 = ( 8.05502,  9.66459)
        z_2 = (-8.05502, -9.66459)
        z_3 = (-8.05502,  9.66459)
        z_4 = ( 8.05502, -9.66459)
    """

    dim = 2
    continuous_inds = list(range(dim))
    _bounds = [(-10.0, 10.0), (-10.0, 10.0)]
    _optimal_value = -19.2085
    _optimizers = [
        (8.05502, 9.66459),
        (-8.05502, -9.66459),
        (-8.05502, 9.66459),
        (8.05502, -9.66459),
    ]

    def _evaluate_true(self, X: Tensor) -> Tensor:
        term = torch.abs(1 - torch.linalg.norm(X, dim=-1) / math.pi)
        return -(
            torch.abs(torch.sin(X[..., 0]) * torch.cos(X[..., 1]) * torch.exp(term))
        )


class Levy(SyntheticTestFunction):
    r"""Levy synthetic test function.

    d-dimensional function (usually evaluated on `[-10, 10]^d`):

        f(x) = sin^2(pi w_1) +
            sum_{i=1}^{d-1} (w_i-1)^2 (1 + 10 sin^2(pi w_i + 1)) +
            (w_d - 1)^2 (1 + sin^2(2 pi w_d))

    where `w_i = 1 + (x_i - 1) / 4` for all `i`.

    f has one minimizer for its global minimum at `z_1 = (1, 1, ..., 1)` with
    `f(z_1) = 0`.
    """

    _optimal_value = 0.0

    def __init__(
        self,
        dim=2,
        noise_std: float | None = None,
        negate: bool = False,
        bounds: list[tuple[float, float]] | None = None,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            dim: The (input) dimension.
            noise_std: Standard deviation of the observation noise.
            negate: If True, negate the function.
            bounds: Custom bounds for the function specified as (lower, upper) pairs.
            dtype: The dtype that is used for the bounds of the function.
        """
        self.dim = dim
        self.continuous_inds = list(range(dim))
        if bounds is None:
            bounds = [(-10.0, 10.0) for _ in range(self.dim)]
        self._optimizers = [tuple(1.0 for _ in range(self.dim))]
        super().__init__(noise_std=noise_std, negate=negate, bounds=bounds, dtype=dtype)

    def _evaluate_true(self, X: Tensor) -> Tensor:
        w = 1.0 + (X - 1.0) / 4.0
        part1 = torch.sin(math.pi * w[..., 0]).pow(2)
        part2 = torch.sum(
            (w[..., :-1] - 1.0).pow(2)
            * (1.0 + 10.0 * torch.sin(math.pi * w[..., :-1] + 1.0).pow(2)),
            dim=-1,
        )
        part3 = (w[..., -1] - 1.0).pow(2) * (
            1.0 + torch.sin(2.0 * math.pi * w[..., -1]).pow(2)
        )
        return part1 + part2 + part3


class Michalewicz(SyntheticTestFunction):
    r"""Michalewicz synthetic test function.

    d-dim function (usually evaluated on hypercube [0, pi]^d):

        M(x) = sum_{i=1}^d sin(x_i) (sin(i x_i^2 / pi)^20)
    """

    def __init__(
        self,
        dim=2,
        noise_std: float | None = None,
        negate: bool = False,
        bounds: list[tuple[float, float]] | None = None,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            dim: The (input) dimension.
            noise_std: Standard deviation of the observation noise.
            negate: If True, negate the function.
            bounds: Custom bounds for the function specified as (lower, upper) pairs.
        """
        self.dim = dim
        self.continuous_inds = list(range(dim))
        if bounds is None:
            bounds = [(0.0, math.pi) for _ in range(self.dim)]
        optvals = {2: -1.80130341, 5: -4.687658, 10: -9.66015}
        optimizers = {2: [(2.20290552, 1.57079633)]}
        self._optimal_value = optvals.get(self.dim)
        self._optimizers = optimizers.get(self.dim)
        super().__init__(noise_std=noise_std, negate=negate, bounds=bounds, dtype=dtype)
        self.register_buffer(
            "i", torch.tensor(tuple(range(1, self.dim + 1)), dtype=self.bounds.dtype)
        )

    @property
    def optimizers(self) -> Tensor:
        if self.dim in (5, 10):
            raise NotImplementedError()
        return super().optimizers

    def _evaluate_true(self, X: Tensor) -> Tensor:
        self.to(device=X.device, dtype=X.dtype)
        m = 10
        return -(
            torch.sum(
                torch.sin(X) * torch.sin(self.i * X.pow(2) / math.pi).pow(2 * m), dim=-1
            )
        )


class Powell(SyntheticTestFunction):
    r"""Powell synthetic test function.

    `d`-dim function (usually evaluated on the hypercube `[-4, 5]^d`):

        P(x) = sum_{i=1}^d/4 (
        (x_{4i-3} + 10 x_{4i-2})**2
        + 5 (x_{4i-1} - x_{4i})**2
        + (x_{4i-2} - 2 x_{4i-1})**4
        + 10 (x_{4i-3} - x_{4i})**4
        )


    P has a global minimizer at `z = (0, ..., 0)` with `P(z) = 0`.
    """

    _optimal_value = 0.0

    def __init__(
        self,
        dim=4,
        noise_std: float | None = None,
        negate: bool = False,
        bounds: list[tuple[float, float]] | None = None,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            dim: The (input) dimension.
            noise_std: Standard deviation of the observation noise.
            negate: If True, negate the function.
            bounds: Custom bounds for the function specified as (lower, upper) pairs.
            dtype: The dtype that is used for the bounds of the function.
        """
        self.dim = dim
        self.continuous_inds = list(range(dim))
        if bounds is None:
            bounds = [(-4.0, 5.0) for _ in range(self.dim)]
        self._optimizers = [tuple(0.0 for _ in range(self.dim))]
        super().__init__(noise_std=noise_std, negate=negate, bounds=bounds, dtype=dtype)

    def _evaluate_true(self, X: Tensor) -> Tensor:
        result = torch.zeros_like(X[..., 0])
        for i in range(self.dim // 4):
            i_ = i + 1
            part1 = (X[..., 4 * i_ - 4] + 10.0 * X[..., 4 * i_ - 3]).pow(2)
            part2 = 5.0 * (X[..., 4 * i_ - 2] - X[..., 4 * i_ - 1]).pow(2)
            part3 = (X[..., 4 * i_ - 3] - 2.0 * X[..., 4 * i_ - 2]).pow(4)
            part4 = 10.0 * (X[..., 4 * i_ - 4] - X[..., 4 * i_ - 1]).pow(4)
            result += part1 + part2 + part3 + part4
        return result


class Rastrigin(SyntheticTestFunction):
    _optimal_value = 0.0

    def __init__(
        self,
        dim=2,
        noise_std: float | None = None,
        negate: bool = False,
        bounds: list[tuple[float, float]] | None = None,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            dim: The (input) dimension.
            noise_std: Standard deviation of the observation noise.
            negate: If True, negate the function.
            bounds: Custom bounds for the function specified as (lower, upper) pairs.
            dtype: The dtype that is used for the bounds of the function.
        """
        self.dim = dim
        self.continuous_inds = list(range(dim))
        if bounds is None:
            bounds = [(-5.12, 5.12) for _ in range(self.dim)]
        self._optimizers = [tuple(0.0 for _ in range(self.dim))]
        super().__init__(noise_std=noise_std, negate=negate, bounds=bounds, dtype=dtype)

    def _evaluate_true(self, X: Tensor) -> Tensor:
        return 10.0 * self.dim + torch.sum(
            X.pow(2) - 10.0 * torch.cos(2.0 * math.pi * X), dim=-1
        )


class Rosenbrock(SyntheticTestFunction):
    r"""Rosenbrock synthetic test function.

    d-dimensional function (usually evaluated on `[-5, 10]^d`):

        f(x) = sum_{i=1}^{d-1} (100 (x_{i+1} - x_i^2)^2 + (x_i - 1)^2)

    f has one minimizer for its global minimum at `z_1 = (1, 1, ..., 1)` with
    `f(z_i) = 0.0`.
    """

    _optimal_value = 0.0

    def __init__(
        self,
        dim=2,
        noise_std: float | None = None,
        negate: bool = False,
        bounds: list[tuple[float, float]] | None = None,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            dim: The (input) dimension.
            noise_std: Standard deviation of the observation noise.
            negate: If True, negate the function.
            bounds: Custom bounds for the function specified as (lower, upper) pairs.
            dtype: The dtype that is used for the bounds of the function.
        """
        self.dim = dim
        self.continuous_inds = list(range(dim))
        if bounds is None:
            bounds = [(-5.0, 10.0) for _ in range(self.dim)]
        self._optimizers = [tuple(1.0 for _ in range(self.dim))]
        super().__init__(noise_std=noise_std, negate=negate, bounds=bounds, dtype=dtype)

    def _evaluate_true(self, X: Tensor) -> Tensor:
        return torch.sum(
            100.0 * (X[..., 1:] - X[..., :-1].pow(2)).pow(2) + (X[..., :-1] - 1).pow(2),
            dim=-1,
        )


class Shekel(SyntheticTestFunction):
    r"""Shekel synthtetic test function.

    4-dimensional function (usually evaluated on `[0, 10]^4`):

        f(x) = -sum_{i=1}^10 (sum_{j=1}^4 (x_j - A_{ji})^2 + C_i)^{-1}

    f has one minimizer for its global minimum at `z_1 = (4, 4, 4, 4)` with
    `f(z_1) = -10.5363`.
    """

    dim = 4
    continuous_inds = list(range(dim))
    _bounds = [(0.0, 10.0), (0.0, 10.0), (0.0, 10.0), (0.0, 10.0)]
    _optimizers = [(4.000747, 3.99951, 4.00075, 3.99951)]

    def __init__(
        self,
        m: int = 10,
        noise_std: float | None = None,
        negate: bool = False,
        bounds: list[tuple[float, float]] | None = None,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            m: Defaults to 10.
            noise_std: Standard deviation of the observation noise.
            negate: If True, negate the function.
            bounds: Custom bounds for the function specified as (lower, upper) pairs.
            dtype: The dtype that is used for the bounds of the function.
        """
        self.m = m
        optvals = {5: -10.1532, 7: -10.4029, 10: -10.536443}
        self._optimal_value = optvals[self.m]
        super().__init__(noise_std=noise_std, negate=negate, bounds=bounds, dtype=dtype)

        self.register_buffer("beta", torch.tensor([1, 2, 2, 4, 4, 6, 3, 7, 5, 5.0]))
        C_t = torch.tensor(
            [
                [4, 1, 8, 6, 3, 2, 5, 8, 6, 7],
                [4, 1, 8, 6, 7, 9, 3, 1, 2, 3.6],
                [4, 1, 8, 6, 3, 2, 5, 8, 6, 7],
                [4, 1, 8, 6, 7, 9, 3, 1, 2, 3.6],
            ],
        )
        self.register_buffer("C", C_t.transpose(-1, -2))

    def _evaluate_true(self, X: Tensor) -> Tensor:
        self.to(device=X.device, dtype=X.dtype)
        beta = self.beta / 10.0
        result = -sum(
            1 / (torch.sum((X - self.C[i]).pow(2), dim=-1) + beta[i])
            for i in range(self.m)
        )
        return result


class SixHumpCamel(SyntheticTestFunction):
    dim = 2
    continuous_inds = list(range(dim))
    _bounds = [(-3.0, 3.0), (-2.0, 2.0)]
    _optimal_value = -1.0316
    _optimizers = [(0.0898, -0.7126), (-0.0898, 0.7126)]

    def _evaluate_true(self, X: Tensor) -> Tensor:
        x1, x2 = X[..., 0], X[..., 1]
        return (
            (4 - 2.1 * x1.pow(2) + x1.pow(4) / 3) * x1.pow(2)
            + x1 * x2
            + (4 * x2.pow(2) - 4) * x2.pow(2)
        )


class StyblinskiTang(SyntheticTestFunction):
    r"""Styblinski-Tang synthtetic test function.

    d-dimensional function (usually evaluated on the hypercube `[-5, 5]^d`):

        H(x) = 0.5 * sum_{i=1}^d (x_i^4 - 16 * x_i^2 + 5 * x_i)

    H has a single global mininimum `H(z) = -39.166166 * d` at `z = [-2.903534]^d`
    """

    def __init__(
        self,
        dim=2,
        noise_std: float | None = None,
        negate: bool = False,
        bounds: list[tuple[float, float]] | None = None,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            dim: The (input) dimension.
            noise_std: Standard deviation of the observation noise.
            negate: If True, negate the function.
            bounds: Custom bounds for the function specified as (lower, upper) pairs.
            dtype: The dtype that is used for the bounds of the function.
        """
        self.dim = dim
        self.continuous_inds = list(range(dim))
        if bounds is None:
            bounds = [(-5.0, 5.0) for _ in range(self.dim)]
        self._optimal_value = -39.166166 * self.dim
        self._optimizers = [tuple(-2.903534 for _ in range(self.dim))]
        super().__init__(noise_std=noise_std, negate=negate, bounds=bounds, dtype=dtype)

    def _evaluate_true(self, X: Tensor) -> Tensor:
        return 0.5 * (X.pow(4) - 16 * X.pow(2) + 5 * X).sum(dim=-1)


class ThreeHumpCamel(SyntheticTestFunction):
    dim = 2
    continuous_inds = list(range(dim))
    _bounds = [(-5.0, 5.0), (-5.0, 5.0)]
    _optimal_value = 0.0
    _optimizers = [(0.0, 0.0)]

    def _evaluate_true(self, X: Tensor) -> Tensor:
        x1, x2 = X[..., 0], X[..., 1]
        return (
            2.0 * x1.pow(2) - 1.05 * x1.pow(4) + x1.pow(6) / 6.0 + x1 * x2 + x2.pow(2)
        )


class AckleyMixed(SyntheticTestFunction):
    r"""Mixed search space version of the Ackley problem.

    This problem has dim-3 binary parameters and 3 continuous parameters in the
    range [0, 1]. This means dim > 3 is required. To make the problem a bit more
    interesting, the optimal value is not at the origin, but rather at `x_opt`
    which is a randomly generated point.

    The goal is  to minimize f(x) = Ackley(x - x_opt).
    """

    _optimal_value = 0.0

    def __init__(
        self,
        dim=53,
        noise_std: float | None = None,
        negate: bool = False,
        randomize_optimum: bool = False,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            dim: The (input) dimension. Must be > 3.
            noise_std: Standard deviation of the observation noise.
            negate: If True, negate the function.
            randomize_optimum: If True, the optimum is a random point in the domain.
            dtype: The dtype that is used for the bounds of the function.
        """
        if dim <= 3:
            raise ValueError(f"Expected dim > 3. Got {dim=}.")
        if randomize_optimum:
            x_opt = torch.rand(dim, dtype=dtype)
            x_opt[: dim - 3] = x_opt[: dim - 3].round()
        else:
            x_opt = torch.zeros(dim, dtype=dtype)
        self._optimizers = [tuple(x.item() for x in x_opt)]
        self.dim = dim
        self.discrete_inds = list(range(0, dim - 3))
        self.continuous_inds = list(range(dim - 3, dim))
        bounds = [(0.0, 1.0) for _ in range(self.dim)]
        super().__init__(bounds=bounds, dtype=dtype, noise_std=noise_std, negate=negate)
        self.register_buffer("x_opt", x_opt)
        self._ackley = Ackley(dim=dim, dtype=dtype)

    def _evaluate_true(self, X: Tensor) -> Tensor:
        return self._ackley.evaluate_true((X - self.x_opt).abs())


class Labs(SyntheticTestFunction):
    r"""Low Auto-correlation Binary Sequences (LABS) problem.

    This input space is binary and the goal is to maximize the Merit factor.
    [Packebusch2016]_ provides optimal values and optimizers attained through
    brute-force for dim <= 66. We include these for dim = 10, 20, 30, 40, 50, 60.
    """

    _check_grad_at_opt = False
    _is_minimization_by_default = False

    def __init__(
        self,
        dim: int = 30,
        noise_std: float | None = None,
        negate: bool = False,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            dim: The (input) dimension.
            noise_std: Standard deviation of the observation noise.
            negate: If True, negate the function.
            bounds: Custom bounds for the function specified as (lower, upper) pairs.
            dtype: The dtype that is used for the bounds of the function.
        """
        self.dim = dim
        self.discrete_inds = list(range(dim))
        bounds = [(0.0, 1.0) for _ in range(dim)]
        optvals = {10: 3.846, 20: 7.692, 30: 7.627, 40: 7.407, 50: 8.170, 60: 8.257}
        optimizers = {
            10: "42211",
            20: "5113112321",
            30: "551212111113231",
            40: "44412112131121313131",
            50: "215131311224112241141142",
            60: "761112141111131124211322211222",
        }
        self._optimal_value = optvals.get(self.dim)
        _optimizers = optimizers.get(self.dim)
        if _optimizers is not None:
            _optimizers = self._optimizer_from_binary_seq(_optimizers)
        self._optimizers = _optimizers
        super().__init__(noise_std=noise_std, negate=negate, bounds=bounds, dtype=dtype)

    def _optimizer_from_binary_seq(self, seq: str) -> list[tuple[float]]:
        """Converts a binary sequence into a an array."""
        arr, val = [], 0
        for s in seq:
            arr += [val for _ in range(int(s))]
            val = 1 - val  # alternate between 0 and 1
        return [tuple(arr)]

    def _evaluate_true(self, X: Tensor) -> Tensor:
        X = 2 * X - 1  # Map from {0, 1}^d to {-1, 1}^d
        energy = torch.zeros(X.shape[:-1], dtype=X.dtype, device=X.device)
        for k in range(1, self.dim):
            energy += (
                (X[..., 0 : self.dim - k] * X[..., k : self.dim]).sum(dim=-1).pow(2)
            )
        return (self.dim**2) / (2.0 * energy)


#  ------------ Constrained synthetic test functions ----------- #


class ConstrainedSyntheticTestFunction(
    ConstrainedBaseTestProblem, SyntheticTestFunction, ABC
):
    r"""Base class for constrained synthetic test functions."""

    def __init__(
        self,
        noise_std: None | float | list[float] = None,
        constraint_noise_std: None | float | list[float] = None,
        negate: bool = False,
        bounds: list[tuple[float, float]] | None = None,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            noise_std: Standard deviation of the observation noise. If a list is
                provided, specifies separate noise standard deviations for each
                objective in a multiobjective problem.
            constraint_noise_std: Standard deviation of the constraint noise.
                If a list is provided, specifies separate noise standard
                deviations for each constraint.
            negate: If True, negate the function.
            bounds: Custom bounds for the function specified as (lower, upper) pairs.
            dtype: The dtype that is used for the bounds of the function.
        """
        SyntheticTestFunction.__init__(
            self, noise_std=noise_std, negate=negate, bounds=bounds, dtype=dtype
        )
        self.constraint_noise_std = self._validate_constraint_noise(
            constraint_noise_std
        )

    def _validate_constraint_noise(
        self, constraint_noise_std
    ) -> None | float | list[float]:
        """
        Validates that constraint_noise_std has length equal to
        the number of constraints, if given as a list

        Args:
            constraint_noise_std: Standard deviation of the constraint noise.
                If a list is provided, specifies separate noise standard
                deviations for each constraint.
        """
        if (
            isinstance(constraint_noise_std, list)
            and len(constraint_noise_std) != self.num_constraints
        ):
            raise InputDataError(
                "If specified as a list, length of constraint_noise_std "
                f"({len(constraint_noise_std)}) must match the "
                f"number of constraints ({self.num_constraints})"
            )
        return constraint_noise_std


class ConstrainedGramacy(ConstrainedSyntheticTestFunction):
    r"""Constrained Gramacy test function.

    This problem comes from [Gramacy2016]_. The problem is defined
    over the unit cube and the goal is to minimize x1+x2 subject to
    1.5 - x1 - 2 * x2 - 0.5 * sin(2*pi*(x1^2 - 2 * x2)) <= 0
    and x1^2 + x2^2 - 1.5 <= 0.
    """

    num_objectives = 1
    num_constraints = 2
    dim = 2
    continuous_inds = list(range(dim))
    _bounds = [(0.0, 1.0), (0.0, 1.0)]
    _optimizers = [(0.1954, 0.4044)]
    _optimal_value = 0.5998  # approximate from [Gramacy2016]_

    def _evaluate_true(self, X: Tensor) -> Tensor:
        """
        Evaluate the function (w/o observation noise) on a set of points.

        Args:
            X: A `batch_shape x d`-dim tensor of point(s) at which to evaluate the
                function.
        """
        return X.sum(dim=-1)

    def _evaluate_slack_true(self, X: Tensor) -> Tensor:
        x1, x2 = X.split(1, dim=-1)
        c1 = 1.5 - x1 - 2 * x2 - 0.5 * torch.sin(2 * math.pi * (x1.pow(2) - 2 * x2))
        c2 = x1.pow(2) + x2.pow(2) - 1.5
        return torch.cat([-c1, -c2], dim=-1)


class ConstrainedHartmann(Hartmann, ConstrainedSyntheticTestFunction):
    r"""Constrained Hartmann test function.

    This is a constrained version of the standard Hartmann test function that
    uses `||x||_2 <= 1` as the constraint. This problem comes from [Letham2019]_.
    """

    num_constraints = 1

    def __init__(
        self,
        dim: int = 6,
        noise_std: None | float = None,
        constraint_noise_std: None | float | list[float] = None,
        negate: bool = False,
        bounds: list[tuple[float, float]] | None = None,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            dim: The (input) dimension.
            noise_std: Standard deviation of the observation noise.
            constraint_noise_std: Standard deviation of the constraint noise.
                If a list is provided, specifies separate noise standard
                deviations for each constraint.
            negate: If True, negate the function.
            bounds: Custom bounds for the function specified as (lower, upper) pairs.
            dtype: The dtype that is used for the bounds of the function.
        """
        Hartmann.__init__(
            self,
            dim=dim,
            noise_std=noise_std,
            negate=negate,
            bounds=bounds,
            dtype=dtype,
        )
        self.constraint_noise_std = self._validate_constraint_noise(
            constraint_noise_std
        )

    def _evaluate_slack_true(self, X: Tensor) -> Tensor:
        return -X.norm(dim=-1, keepdim=True) + 1


class ConstrainedHartmannSmooth(Hartmann, ConstrainedSyntheticTestFunction):
    r"""Smooth constrained Hartmann test function.

    This is a constrained version of the standard Hartmann test function that
    uses `||x||_2^2 <= 1` as the constraint to obtain smoother constraint slack.
    """

    num_constraints = 1

    def __init__(
        self,
        dim: int = 6,
        noise_std: None | float = None,
        constraint_noise_std: None | float | list[float] = None,
        negate: bool = False,
        bounds: list[tuple[float, float]] | None = None,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            dim: The (input) dimension.
            noise_std: Standard deviation of the observation noise.
            constraint_noise_std: Standard deviation of the constraint noise.
                If a list is provided, specifies separate noise standard
                deviations for each constraint.
            negate: If True, negate the function.
            bounds: Custom bounds for the function specified as (lower, upper) pairs.
            dtype: The dtype that is used for the bounds of the function.
        """
        Hartmann.__init__(
            self,
            dim=dim,
            noise_std=noise_std,
            negate=negate,
            bounds=bounds,
            dtype=dtype,
        )
        self.constraint_noise_std = self._validate_constraint_noise(
            constraint_noise_std
        )

    def _evaluate_slack_true(self, X: Tensor) -> Tensor:
        return -X.pow(2).sum(dim=-1, keepdim=True) + 1


class PressureVessel(ConstrainedSyntheticTestFunction):
    r"""Pressure vessel design problem with constraints.

    The four-dimensional pressure vessel design problem with four black-box
    constraints from [CoelloCoello2002constraint]_.
    """

    dim = 4
    continuous_inds = list(range(dim))
    num_constraints = 4
    _bounds = [(0.0, 10.0), (0.0, 10.0), (10.0, 50.0), (150.0, 200.0)]
    _optimal_value = 6059.946341  # from [CoelloCoello2002constraint]

    def _evaluate_true(self, X: Tensor) -> Tensor:
        x1, x2, x3, x4 = X.unbind(-1)
        x1 = round_nearest(x1, increment=0.0625, bounds=self._bounds[0])
        x2 = round_nearest(x2, increment=0.0625, bounds=self._bounds[1])
        return (
            0.6224 * x1 * x3 * x4
            + 1.7781 * x2 * x3.pow(2)
            + 3.1661 * x1.pow(2) * x4
            + 19.84 * x1.pow(2) * x3
        )

    def _evaluate_slack_true(self, X: Tensor) -> Tensor:
        x1, x2, x3, x4 = X.unbind(-1)
        return -torch.stack(
            [
                -x1 + 0.0193 * x3,
                -x2 + 0.00954 * x3,
                -math.pi * x3.pow(2) * x4 - (4 / 3) * math.pi * x3.pow(3) + 1296000.0,
                x4 - 240.0,
            ],
            dim=-1,
        )


class WeldedBeamSO(ConstrainedSyntheticTestFunction):
    r"""Welded beam design problem with constraints (single-outcome).

    The four-dimensional welded beam design proble problem with six
    black-box constraints from [CoelloCoello2002constraint]_.

    For a (somewhat modified) multi-objective version, see
    `botorch.test_functions.multi_objective.WeldedBeam`.
    """

    dim = 4
    continuous_inds = list(range(dim))
    num_constraints = 6
    _bounds = [(0.125, 10.0), (0.1, 10.0), (0.1, 10.0), (0.1, 10.0)]
    _optimal_value = 1.728226  # from [CoelloCoello2002constraint]

    def _evaluate_true(self, X: Tensor) -> Tensor:
        x1, x2, x3, x4 = X.unbind(-1)
        return 1.10471 * x1.pow(2) * x2 + 0.04811 * x3 * x4 * (14.0 + x2)

    def _evaluate_slack_true(self, X: Tensor) -> Tensor:
        x1, x2, x3, x4 = X.unbind(-1)
        P = 6000.0
        L = 14.0
        E = 30e6
        G = 12e6
        t_max = 13600.0
        s_max = 30000.0
        d_max = 0.25

        M = P * (L + x2 / 2)
        R = torch.sqrt(0.25 * (x2.pow(2) + (x1 + x3).pow(2)))
        J = 2 * math.sqrt(2) * x1 * x2 * (x2.pow(2) / 12 + 0.25 * (x1 + x3).pow(2))
        P_c = (
            4.013
            * E
            * x3
            * x4.pow(3)
            * 6
            / (L**2)
            * (1 - 0.25 * x3 * math.sqrt(E / G) / L)
        )
        t1 = P / (math.sqrt(2) * x1 * x2)
        t2 = M * R / J
        t = torch.sqrt(t1.pow(2) + t1 * t2 * x2 / R + t2.pow(2))
        s = 6 * P * L / (x4 * x3.pow(2))
        d = 4 * P * L**3 / (E * x3.pow(3) * x4)

        g1 = t - t_max
        g2 = s - s_max
        g3 = x1 - x4
        g4 = 0.10471 * x1.pow(2) + 0.04811 * x3 * x4 * (14.0 + x2) - 5.0
        g5 = d - d_max
        g6 = P - P_c

        return -torch.stack([g1, g2, g3, g4, g5, g6], dim=-1)


class TensionCompressionString(ConstrainedSyntheticTestFunction):
    r"""Tension compression string optimization problem with constraints.

    The three-dimensional tension compression string optimization problem with
    four black-box constraints from [Hedar2006derivfree]_.
    """

    dim = 3
    continuous_inds = list(range(dim))
    num_constraints = 4
    _bounds = [(0.01, 1.0), (0.01, 1.0), (0.01, 20.0)]
    _optimal_value = 0.012681  # from [CoelloCoello2002constraint]

    def _evaluate_true(self, X: Tensor) -> Tensor:
        x1, x2, x3 = X.unbind(-1)
        return x1.pow(2) * x2 * (x3 + 2)

    def _evaluate_slack_true(self, X: Tensor) -> Tensor:
        x1, x2, x3 = X.unbind(-1)
        constraints = torch.stack(
            [
                1 - x2.pow(3) * x3 / (71785 * x1.pow(4)),
                (4 * x2.pow(2) - x1 * x2) / (12566 * x1.pow(3) * (x2 - x1))
                + 1 / (5108 * x1.pow(2))
                - 1,
                1 - 140.45 * x1 / (x3 * x2.pow(2)),
                (x1 + x2) / 1.5 - 1,
            ],
            dim=-1,
        )
        return -constraints.clamp_max(100)


class SpeedReducer(ConstrainedSyntheticTestFunction):
    r"""Speed Reducer design problem with constraints.

    The seven-dimensional speed reducer design problem with eleven black-box
    constraints from [Lemonge2010constrained]_.
    """

    dim = 7
    continuous_inds = list(range(dim))
    num_constraints = 11
    _bounds = [
        (2.6, 3.6),
        (0.7, 0.8),
        (17.0, 28.0),
        (7.3, 8.3),
        (7.8, 8.3),
        (2.9, 3.9),
        (5.0, 5.5),
    ]
    _optimal_value = 2996.3482  # from [Lemonge2010constrained]

    def _evaluate_true(self, X: Tensor) -> Tensor:
        x1, x2, x3, x4, x5, x6, x7 = X.unbind(-1)
        return (
            0.7854 * x1 * x2.pow(2) * (3.3333 * x3.pow(2) + 14.9334 * x3 - 43.0934)
            + -1.508 * x1 * (x6.pow(2) + x7.pow(2))
            + 7.4777 * (x6.pow(3) + x7.pow(3))
            + 0.7854 * (x4 * x6.pow(2) + x5 * x7.pow(2))
        )

    def _evaluate_slack_true(self, X: Tensor) -> Tensor:
        x1, x2, x3, x4, x5, x6, x7 = X.unbind(-1)
        return -torch.stack(
            [
                27.0 * (1 / x1) * (1 / x2.pow(2)) * (1 / x3) - 1,
                397.5 * (1 / x1) * (1 / x2.pow(2)) * (1 / x3.pow(2)) - 1,
                1.93 * (1 / x2) * (1 / x3) * x4.pow(3) * (1 / x6.pow(4)) - 1,
                1.93 * (1 / x2) * (1 / x3) * x5.pow(3) * (1 / x7.pow(4)) - 1,
                1
                / (0.1 * x6.pow(3))
                * torch.sqrt((745 * x4 / (x2 * x3)).pow(2) + 16.9 * 1e6)
                - 1100,
                1
                / (0.1 * x7.pow(3))
                * torch.sqrt((745 * x5 / (x2 * x3)).pow(2) + 157.5 * 1e6)
                - 850,
                x2 * x3 - 40,
                5 - x1 / x2,
                x1 / x2 - 12,
                (1.5 * x6 + 1.9) / x4 - 1,
                (1.1 * x7 + 1.9) / x5 - 1,
            ],
            dim=-1,
        )


class KeaneBumpFunction(ConstrainedSyntheticTestFunction):
    r"""Keane Bump Function problem with constraints.

    This is a challenging d-dimensional minimization problem with two
    constraints that is evaluated on the domain [0, 10]^d.

    There is no known global optimum for this problem, but the aproximate
    global optimal value for a few different dimensionalities can be found in
    [Mishra2007]_.
    """

    num_constraints = 2
    _optimal_value_lookup = {
        2: -0.365,
        10: -0.6737,
        15: -0.781647601,
        20: -0.803619104,
        30: -0.818056222,
        40: -0.826624404,
        50: -0.83078783,
    }

    def __init__(
        self,
        dim: int,
        noise_std: None | float = None,
        constraint_noise_std: None | float | list[float] = None,
        negate: bool = False,
        bounds: list[tuple[float, float]] | None = None,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""
        Args:
            dim: The (input) dimension.
            noise_std: Standard deviation of the observation noise.
            constraint_noise_std: Standard deviation of the constraint noise.
                If a list is provided, specifies separate noise standard
                deviations for each constraint.
            negate: If True, negate the function.
            bounds: Custom bounds for the function specified as (lower, upper) pairs.
            dtype: The dtype that is used for the bounds of the function.
        """
        self.dim = dim
        self.continuous_inds = list(range(dim))
        if bounds is None:
            bounds = [(0.0, 10.0) for _ in range(dim)]
        if dim in self._optimal_value_lookup.keys():
            self._optimal_value = self._optimal_value_lookup[dim]
        super().__init__(
            noise_std=noise_std,
            constraint_noise_std=constraint_noise_std,
            negate=negate,
            bounds=bounds,
            dtype=dtype,
        )

    def _evaluate_true(self, X: Tensor) -> Tensor:
        Xcos = X.cos()
        num = Xcos.pow(4).sum(dim=-1) - 2 * Xcos.pow(2).prod(dim=-1)
        den = torch.sqrt(
            (torch.arange(1, self.dim + 1, device=X.device) * X.pow(2)).sum(dim=-1)
        )
        # clamp to avoid den=0, which happens when X=0.
        return -(num / den.clamp(min=1e-3)).abs()

    def _evaluate_slack_true(self, X: Tensor) -> Tensor:
        return torch.cat(
            [
                X.prod(dim=-1, keepdims=True) - 0.75,
                7.5 * self.dim - X.sum(dim=-1, keepdims=True),
            ],
            dim=-1,
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any

from botorch.models.gp_regression import SingleTaskGP
from botorch.models.kernels.contextual_lcea import LCEAKernel
from botorch.models.kernels.contextual_sac import SACKernel
from botorch.utils.datasets import SupervisedDataset
from torch import Tensor


class SACGP(SingleTaskGP):
    r"""A GP using a Structural Additive Contextual(SAC) kernel."""

    def __init__(
        self,
        train_X: Tensor,
        train_Y: Tensor,
        train_Yvar: Tensor | None,
        decomposition: dict[str, list[int]],
    ) -> None:
        r"""
        Args:
            train_X: (n x d) X training data.
            train_Y: (n x 1) Y training data.
            train_Yvar: (n x 1) Noise variances of each training Y. If None,
                we use an inferred noise likelihood.
            decomposition: Keys are context names. Values are the indexes of
                parameters belong to the context. The parameter indexes are in
                the same order across contexts.
        """
        super().__init__(train_X=train_X, train_Y=train_Y, train_Yvar=train_Yvar)
        self.covar_module = SACKernel(
            decomposition=decomposition,
            batch_shape=self._aug_batch_shape,
            device=train_X.device,
        )
        self.decomposition = decomposition
        self.to(train_X)

    @classmethod
    def construct_inputs(
        cls,
        training_data: SupervisedDataset,
        decomposition: dict[str, list[int]],
    ) -> dict[str, Any]:
        r"""Construct `Model` keyword arguments from a dict of `SupervisedDataset`.

        Args:
            training_data: A `SupervisedDataset` containing the training data.
            decomposition: Dictionary of context names and their indexes of the
                corresponding active context parameters.
        """
        base_inputs = super().construct_inputs(training_data=training_data)
        return {
            **base_inputs,
            "decomposition": decomposition,
        }


class LCEAGP(SingleTaskGP):
    r"""A GP using a Latent Context Embedding Additive (LCE-A) Kernel.

    Note that the model does not support batch training. Input training
    data sets should have dim = 2.
    """

    def __init__(
        self,
        train_X: Tensor,
        train_Y: Tensor,
        train_Yvar: Tensor | None,
        decomposition: dict[str, list[int]],
        train_embedding: bool = True,
        cat_feature_dict: dict | None = None,
        embs_feature_dict: dict | None = None,
        embs_dim_list: list[int] | None = None,
        context_weight_dict: dict | None = None,
    ) -> None:
        r"""
        Args:
            train_X: (n x d) X training data.
            train_Y: (n x 1) Y training data.
            train_Yvar: (n x 1) Noise variance of Y. If None,
                we use an inferred noise likelihood.
            decomposition: Keys are context names. Values are the indexes of
                parameters belong to the context.
            train_embedding: Whether to train the embedding layer or not. If False,
                the model will use pre-trained embeddings in embs_feature_dict.
            cat_feature_dict: Keys are context names and values are list of categorical
                features i.e. {"context_name" : [cat_0, ..., cat_k]}, where k is the
                number of categorical variables. If None, we use context names in the
                decomposition as the only categorical feature, i.e., k = 1.
            embs_feature_dict: Pre-trained continuous embedding features of each
                context.
            embs_dim_list: Embedding dimension for each categorical variable. The length
                equals the number of categorical features k. If None, the embedding
                dimension is set to 1 for each categorical variable.
            context_weight_dict: Known population weights of each context.
        """
        super().__init__(
            train_X=train_X,
            train_Y=train_Y,
            train_Yvar=train_Yvar,
            outcome_transform=None,
        )
        self.covar_module = LCEAKernel(
            decomposition=decomposition,
            batch_shape=self._aug_batch_shape,
            train_embedding=train_embedding,
            cat_feature_dict=cat_feature_dict,
            embs_feature_dict=embs_feature_dict,
            embs_dim_list=embs_dim_list,
            context_weight_dict=context_weight_dict,
            device=train_X.device,
        )
        self.decomposition = decomposition
        self.to(train_X)

    @classmethod
    def construct_inputs(
        cls,
        training_data: SupervisedDataset,
        decomposition: dict[str, list[str]],
        train_embedding: bool = True,
        cat_feature_dict: dict | None = None,
        embs_feature_dict: dict | None = None,
        embs_dim_list: list[int] | None = None,
        context_weight_dict: dict | None = None,
    ) -> dict[str, Any]:
        r"""Construct `Model` keyword arguments from a dict of `SupervisedDataset`.

        Args:
            training_data: A `SupervisedDataset` containing the training data.
            decomposition: Dictionary of context names and the names of the
                corresponding active context parameters.
            train_embedding: Whether to train the embedding layer or not.
            cat_feature_dict: Keys are context names and values are list of categorical
                features i.e. {"context_name" : [cat_0, ..., cat_k]}, where k is the
                number of categorical variables. If None, we use context names in the
                decomposition as the only categorical feature, i.e., k = 1.
            embs_feature_dict: Pre-trained continuous embedding features of each
                context.
            embs_dim_list: Embedding dimension for each categorical variable. The length
                equals the number of categorical features k. If None, the embedding
                dimension is set to 1 for each categorical variable.
            context_weight_dict: Known population weights of each context.
        """
        base_inputs = super().construct_inputs(training_data=training_data)
        index_decomp = {
            c: [training_data.feature_names.index(i) for i in v]
            for c, v in decomposition.items()
        }
        return {
            **base_inputs,
            "decomposition": index_decomp,
            "train_embedding": train_embedding,
            "cat_feature_dict": cat_feature_dict,
            "embs_feature_dict": embs_feature_dict,
            "embs_dim_list": embs_dim_list,
            "context_weight_dict": context_weight_dict,
        }



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""Abstract base module for multi-output acquisition functions."""

from __future__ import annotations

from abc import ABC, abstractmethod

import torch
from botorch.acquisition.acquisition import AcquisitionFunction
from botorch.exceptions.errors import UnsupportedError
from botorch.models.model import Model
from botorch.utils.transforms import (
    average_over_ensemble_models,
    t_batch_mode_transform,
)
from torch import Tensor


class MultiOutputAcquisitionFunction(AcquisitionFunction, ABC):
    r"""Abstract base class for multi-output acquisition functions.

    These are intended to be optimized with a multi-objective optimizer (e.g.
    NSGA-II).
    """

    @abstractmethod
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate the acquisition function on the candidate set X.

        Args:
            X: A `(b) x q x d`-dim Tensor of `(b)` t-batches with `q` `d`-dim
                design points each.

        Returns:
            A `(b) x m`-dim Tensor of acquisition function values at the given
            design points `X`.
        """

    def set_X_pending(self, X_pending: Tensor | None) -> None:
        r"""Set the pending points.

        Args:
            X_pending: A `batch_shape x m x d`-dim Tensor of `m` design points that
                have points that have been submitted for function evaluation
                (but may not yet have been evaluated).
        """
        raise UnsupportedError(
            "X_pending is not supported for multi-output acquisition functions."
        )


class MultiOutputPosteriorMean(MultiOutputAcquisitionFunction):
    def __init__(self, model: Model, weights: Tensor | None = None) -> None:
        r"""Constructor for the MultiPosteriorMean.

        Maximization of all outputs is assumed by default. Minimizing outputs can
        be achieved by setting the corresponding weights to negative.

        Args:
            acqfs: A list of `m` acquisition functions.
            weights: A one-dimensional tensor with `m` elements representing the
                weights on the outputs.
        """
        super().__init__(model=model)
        if self.model.num_outputs < 2:
            raise NotImplementedError(
                "MultiPosteriorMean only supports multi-output models."
            )
        # TODO: this could be done via a posterior transform
        if weights is not None and weights.shape[0] != self.model.num_outputs:
            raise ValueError(
                f"weights must have {self.model.num_outputs} elements, but got"
                f" {weights.shape[0]}."
            )
        self.register_buffer("weights", weights)

    @t_batch_mode_transform(expected_q=1, assert_output_shape=False)
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate the acquisition function on the candidate set X.

        Args:
            X: A `(b) x q x d`-dim Tensor of `(b)` t-batches with `q` `d`-dim
                design points each.

        Returns:
            A `(b) x m`-dim Tensor of acquisition function values at the given
            design points `X`.
        """
        mean = self.model.posterior(X).mean.squeeze(-2)
        if self.weights is not None:
            return mean * self.weights
        return mean


class MultiOutputAcquisitionFunctionWrapper(MultiOutputAcquisitionFunction):
    r"""Multi-output wrapper around single-output acquisition functions."""

    def __init__(self, acqfs: list[AcquisitionFunction]) -> None:
        r"""Constructor for the AcquisitionFunction base class.

        Args:
            acqfs: A list of `m` acquisition functions.
        """
        # We could set the model to be an ensemble model consistent of the
        # model used in each acqf
        super().__init__(model=acqfs[0].model)
        self.acqfs: list[AcquisitionFunction] = acqfs

    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate the acquisition function on the candidate set X.

        Args:
            X: A `(b) x q x d`-dim Tensor of `(b)` t-batches with `q` `d`-dim
                design points each.

        Returns:
            A `(b) x m`-dim Tensor of acquisition function values at the given
            design points `X`.
        """
        return torch.stack([acqf(X) for acqf in self.acqfs], dim=-1)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import torch
from botorch.acquisition.analytic import AcquisitionFunction
from botorch.acquisition.objective import (
    IdentityMCObjective,
    MCAcquisitionObjective,
    PosteriorTransform,
)
from botorch.exceptions.errors import UnsupportedError
from botorch.models.deterministic import GenericDeterministicModel
from botorch.models.model import Model
from botorch.sampling.pathwise.posterior_samplers import get_matheron_path_model
from botorch.utils.transforms import is_ensemble, t_batch_mode_transform
from torch import Tensor


BATCH_SIZE_CHANGE_ERROR = """The batch size of PathwiseThompsonSampling should \
not change during a forward pass - was {}, now {}. Please re-initialize the \
acquisition if you want to change the batch size."""


class PathwiseThompsonSampling(AcquisitionFunction):
    r"""Single-outcome Thompson Sampling packaged as an (analytic)
    acquisition function. Querying the acquisition function gives the summed
    values of one or more draws from a pathwise drawn posterior sample, and thus
    it maximization yields one (or multiple) Thompson sample(s).

    Example:
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> TS = PathwiseThompsonSampling(model)
    """

    def __init__(
        self,
        model: Model,
        objective: MCAcquisitionObjective | None = None,
        posterior_transform: PosteriorTransform | None = None,
    ) -> None:
        r"""Single-outcome TS.

        If using a multi-output `model`, the acquisition function requires either an
        `objective` or a `posterior_transform` that transforms the multi-output
        posterior samples to single-output posterior samples.

        Args:
            model: A fitted GP model.
            objective: The MCAcquisitionObjective under which the samples are
                evaluated. Defaults to `IdentityMCObjective()`.
            posterior_transform: An optional PosteriorTransform.
        """

        super().__init__(model=model)
        self.batch_size: int | None = None
        self.samples: GenericDeterministicModel | None = None
        self.ensemble_indices: Tensor | None = None

        # NOTE: This conditional block is copied from MCAcquisitionFunction, we should
        # consider inherting from it and e.g. getting the X_pending logic as well.
        if objective is None and model.num_outputs != 1:
            if posterior_transform is None:
                raise UnsupportedError(
                    "Must specify an objective or a posterior transform when using "
                    "a multi-output model."
                )
            elif not posterior_transform.scalarize:
                raise UnsupportedError(
                    "If using a multi-output model without an objective, "
                    "posterior_transform must scalarize the output."
                )
        if objective is None:
            objective = IdentityMCObjective()
        self.objective = objective
        self.posterior_transform = posterior_transform

    def redraw(self, batch_size: int) -> None:
        sample_shape = (batch_size,)
        self.samples = get_matheron_path_model(
            model=self.model, sample_shape=torch.Size(sample_shape)
        )
        if is_ensemble(self.model):
            # the ensembling dimension is assumed to be part of the batch shape
            model_batch_shape = self.model.batch_shape
            if len(model_batch_shape) > 1:
                raise NotImplementedError(
                    "Ensemble models with more than one ensemble dimension are not "
                    "yet supported."
                )
            num_ensemble = model_batch_shape[0]
            # ensemble_indices is cached here to ensure that the acquisition function
            # becomes deterministic for the same input and can be optimized with LBFGS.
            # ensemble_indices is used in select_from_ensemble_models.
            self.ensemble_indices = torch.randint(
                0,
                num_ensemble,
                (*sample_shape, 1, self.model.num_outputs),
            )

    @t_batch_mode_transform()
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate the pathwise posterior sample draws on the candidate set X.

        Args:
            X: A `batch_shape x q x d`-dim batched tensor of `d`-dim design points.

        Returns:
            A `batch_shape`-dim tensor of evaluations on the posterior sample draws,
            where the samples are summed over the q-batch dimension.
        """
        objective_values = self._pathwise_forward(X)  # batch_shape x q
        # NOTE: The current implementation sums over the q-batch dimension, which means
        # that we are optimizing the sum of independent Thompson samples. In the future,
        # we can leverage *batched* L-BFGS optimization, rather than summing over the q
        # dimension, which will guarantee descent steps for all members of the batch
        # through batch-member-specific learning rate selection.
        return objective_values.sum(-1)  # batch_shape

    def _pathwise_forward(self, X: Tensor) -> Tensor:
        """Evaluate the pathwise posterior sample draws on the candidate set X.

        Args:
            X: A `batch_shape x q x d`-dim batched tensor of `d`-dim design points.

        Returns:
            A `batch_shape x q`-dim tensor of evaluations on the posterior sample draws.
        """
        batch_size = X.shape[-2]
        # batch_shape x q x 1 x d
        X = X.unsqueeze(-2)
        if self.samples is None:
            self.batch_size = batch_size
            self.redraw(batch_size=batch_size)

        if self.batch_size != batch_size:
            raise ValueError(
                BATCH_SIZE_CHANGE_ERROR.format(self.batch_size, batch_size)
            )
        # batch_shape x q [x num_ensembles] x 1 x m
        posterior_values = self.samples(X)
        # batch_shape x q [x num_ensembles] x m
        posterior_values = posterior_values.squeeze(-2)

        # batch_shape x q x m
        posterior_values = self.select_from_ensemble_models(values=posterior_values)

        if self.posterior_transform:
            posterior_values = self.posterior_transform.evaluate(posterior_values)
        # objective removes the `m` dimension
        objective_values = self.objective(posterior_values)  # batch_shape x q
        return objective_values

    def select_from_ensemble_models(self, values: Tensor):
        """Subselecting a value associated with a single sample in the ensemble for each
        element of samples that is not associated with an ensemble dimension.

        NOTE: 1) uses `self.model` and `is_ensemble` to determine whether or not an
        ensembling dimension is present. 2) uses `self.ensemble_indices` to select the
        value associated with a single sample in the ensemble. `ensemble_indices`
        contains uniformly randomly sample indices for each element of the ensemble, but
        is cached to make the evaluation of the acquisition function deterministic.

        Args:
            values: A `batch_shape x num_draws x q [x num_ensemble] x m`-dim Tensor.

        Returns:
            A`batch_shape x num_draws x q x m`-dim where each element is contains a
            single sample from the ensemble, selected with `self.ensemble_indices`.
        """
        if not is_ensemble(self.model):
            return values

        ensemble_dim = -2
        # `ensemble_indices` are fixed so that the acquisition function becomes
        # deterministic for the same input and can be optimized with LBFGS.
        # ensemble indices have shape num_paths x 1 x m
        self.ensemble_indices = self.ensemble_indices.to(device=values.device)
        index = self.ensemble_indices
        input_batch_shape = values.shape[:-3]
        index = index.expand(*input_batch_shape, *index.shape)
        # samples is batch_shape x q x num_ensemble x m
        values_wo_ensemble = torch.gather(values, dim=ensemble_dim, index=index)
        return values_wo_ensemble.squeeze(
            ensemble_dim
        )  # removing the ensemble dimension



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
This file contains a simple modification of SingleTaskGP as an example.

References:

.. [Example2024paper]
    S. Cakmak, D.Eriksson, M. Balandat, E. Bakshy. Example Paper.
    Proceedings of the Example Conference, 2024.

Contributor: saitcakmak
"""

from typing import Optional

from botorch.models.gp_regression import SingleTaskGP
from gpytorch.kernels import RBFKernel, ScaleKernel
from torch import Tensor


class ExampleModel(SingleTaskGP):
    def __init__(
        self, train_X: Tensor, train_Y: Tensor, train_Yvar: Optional[Tensor] = None
    ) -> None:
        r"""Initialize the example model from [Example2024paper]_.

        Args:
            train_X: A `batch_shape x n x d` tensor of training features.
            train_Y: A `batch_shape x n x m` tensor of training observations.
            train_Yvar: An optional `batch_shape x n x m` tensor of observed
                measurement noise.
        """
        super().__init__(
            train_X=train_X,
            train_Y=train_Y,
            train_Yvar=train_Yvar,
            covar_module=ScaleKernel(RBFKernel(ard_num_dims=train_X.shape[-1])),
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import os
from logging import DEBUG, WARN
from unittest.mock import MagicMock, mock_open, patch

import torch
from botorch.acquisition.objective import ScalarizedPosteriorTransform
from botorch.exceptions.errors import UnsupportedError
from botorch.models.transforms.input import Normalize
from botorch.utils.testing import BotorchTestCase
from botorch_community.models.prior_fitted_network import PFNModel
from botorch_community.models.utils.prior_fitted_network import (
    download_model,
    ModelPaths,
)
from torch import nn, Tensor


class DummyPFN(nn.Module):
    def __init__(self, n_buckets: int = 1000):
        """A dummy PFN model for testing purposes.

        This class implements a mocked PFN model that returns
        constant values for testing. It mimics the interface of actual PFN models
        but with simplified behavior.

        Args:
            n_buckets: Number of buckets for the output distribution. Default is 1000.
        """

        super().__init__()
        self.n_buckets = n_buckets
        self.criterion = MagicMock()
        self.criterion.borders = torch.linspace(0, 1, n_buckets + 1)

    def forward(self, train_X: Tensor, train_Y: Tensor, test_X: Tensor) -> Tensor:
        return torch.zeros(*test_X.shape[:-1], self.n_buckets, device=test_X.device)


class TestPriorFittedNetwork(BotorchTestCase):
    def test_raises(self):
        for dtype in (torch.float, torch.double):
            tkwargs = {"device": self.device, "dtype": dtype}
            train_X = torch.rand(10, 3, **tkwargs)
            train_Y = torch.rand(10, 1, **tkwargs)
            train_Yvar = torch.rand(10, 1, **tkwargs)
            test_X = torch.rand(5, 3, **tkwargs)

            with self.assertLogs(logger="botorch", level=DEBUG) as log:
                PFNModel(train_X, train_Y, DummyPFN(), train_Yvar=train_Yvar)
                self.assertIn(
                    "train_Yvar provided but ignored for PFNModel.",
                    log.output[0],
                )

            train_Y_4d = torch.rand(10, 2, 2, 1, **tkwargs)
            with self.assertRaises(UnsupportedError):
                PFNModel(train_X, train_Y_4d, DummyPFN())

            train_Y_2d = torch.rand(10, 2, **tkwargs)
            with self.assertRaises(UnsupportedError):
                PFNModel(train_X, train_Y_2d, DummyPFN())

            with self.assertRaises(UnsupportedError):
                PFNModel(torch.rand(10, 3, 3, 2, **tkwargs), train_Y, DummyPFN())

            with self.assertRaises(UnsupportedError):
                PFNModel(train_X, torch.rand(11, **tkwargs), DummyPFN())

            pfn = PFNModel(train_X, train_Y, DummyPFN())

            with self.assertRaises(RuntimeError):
                pfn.posterior(test_X, output_indices=[0, 1])
            with self.assertLogs(logger="botorch", level=WARN) as log:
                pfn.posterior(test_X, observation_noise=True)
                self.assertIn(
                    "observation_noise is not supported for PFNModel",
                    log.output[0],
                )
            with self.assertRaises(UnsupportedError):
                pfn.posterior(
                    test_X,
                    posterior_transform=ScalarizedPosteriorTransform(
                        weights=torch.ones(1)
                    ),
                )

            # q should be 1
            test_X = torch.rand(5, 4, 2, **tkwargs)
            with self.assertRaises(UnsupportedError):
                pfn.posterior(test_X)

            # X dims should be 1 to 4
            test_X = torch.rand(5, 4, 2, 1, 2, **tkwargs)
            with self.assertRaises(UnsupportedError):
                pfn.posterior(test_X)

    def test_shapes(self):
        tkwargs = {"device": self.device, "dtype": torch.float}

        # no q dimension
        train_X = torch.rand(10, 3, **tkwargs)
        train_Y = torch.rand(10, 1, **tkwargs)
        test_X = torch.rand(5, 3, **tkwargs)

        pfn = PFNModel(train_X, train_Y, DummyPFN(n_buckets=100))

        for batch_first in [True, False]:
            with self.subTest(batch_first=batch_first):
                pfn.batch_first = batch_first
                posterior = pfn.posterior(test_X)

                self.assertEqual(posterior.probabilities.shape, torch.Size([5, 100]))
                self.assertEqual(posterior.mean.shape, torch.Size([5, 1]))

        # q=1
        test_X = torch.rand(5, 1, 3, **tkwargs)
        posterior = pfn.posterior(test_X)

        self.assertEqual(posterior.probabilities.shape, torch.Size([5, 100]))
        self.assertEqual(posterior.mean.shape, torch.Size([5, 1]))

        # no shape basically
        test_X = torch.rand(3, **tkwargs)
        posterior = pfn.posterior(test_X)

        self.assertEqual(posterior.probabilities.shape, torch.Size([100]))
        self.assertEqual(posterior.mean.shape, torch.Size([1]))

    def test_batching(self):
        tkwargs = {"device": self.device, "dtype": torch.float}

        # no q dimension
        train_X = torch.rand(2, 10, 3, **tkwargs)
        train_Y = torch.rand(2, 10, 1, **tkwargs)

        pfn = PFNModel(train_X, train_Y, DummyPFN(n_buckets=100))

        test_X = torch.rand(5, 2, 1, 3, **tkwargs)
        posterior = pfn.posterior(test_X)

        self.assertEqual(posterior.probabilities.shape, torch.Size([5, 2, 100]))

        test_X = torch.rand(2, 1, 3, **tkwargs)
        posterior = pfn.posterior(test_X)

        self.assertEqual(posterior.probabilities.shape, torch.Size([2, 100]))

    def test_input_transform(self):
        model = PFNModel(
            train_X=torch.rand(10, 3),
            train_Y=torch.rand(10, 1),
            input_transform=Normalize(d=3),
            model=DummyPFN(),
        )
        self.assertIsInstance(model.input_transform, Normalize)
        self.assertEqual(model.input_transform.bounds.shape, torch.Size([2, 3]))


class TestPriorFittedNetworkUtils(BotorchTestCase):
    @patch("botorch_community.models.utils.prior_fitted_network.requests.get")
    @patch("botorch_community.models.utils.prior_fitted_network.gzip.GzipFile")
    @patch("botorch_community.models.utils.prior_fitted_network.torch.load")
    @patch("botorch_community.models.utils.prior_fitted_network.torch.save")
    @patch("botorch_community.models.utils.prior_fitted_network.os.path.exists")
    @patch("botorch_community.models.utils.prior_fitted_network.os.makedirs")
    def test_download_model_cache_miss(
        self,
        _mock_makedirs,
        mock_exists,
        mock_torch_save,
        mock_torch_load,
        mock_gzip,
        mock_requests_get,
    ):
        # Simulate cache miss
        mock_exists.return_value = False

        # Mock the requests.get to simulate a network call
        mock_requests_get.return_value = MagicMock(
            status_code=200, content=b"fake content"
        )

        # Mock the gzip.GzipFile to simulate decompression
        mock_gzip.return_value.__enter__.return_value = mock_open(
            read_data=b"fake model data"
        ).return_value

        # Mock torch.load to simulate loading a model
        fake_model = MagicMock(spec=torch.nn.Module)
        mock_torch_load.return_value = fake_model

        # Call the function
        model = download_model(
            ModelPaths.pfns4bo_hebo,
            cache_dir=os.environ.get("RUNNER_TEMP", "/tmp") + "/test_cache",
            # $RUNNER_TEMP is set by GitHub Actions as tmp, /tmp does not work there
        )

        # Assertions for cache miss
        mock_requests_get.assert_called_once()
        mock_gzip.assert_called_once()
        mock_torch_load.assert_called_once()
        mock_torch_save.assert_called_once()
        self.assertEqual(model, fake_model)

        # Test loading in model init
        model = PFNModel(
            train_X=torch.rand(10, 3),
            train_Y=torch.rand(10, 1),
        )
        self.assertEqual(model.pfn, fake_model)

    @patch("botorch_community.models.utils.prior_fitted_network.torch.load")
    @patch("botorch_community.models.utils.prior_fitted_network.os.path.exists")
    def test_download_model_cache_hit(self, mock_exists, mock_torch_load):
        # Simulate cache hit
        mock_exists.return_value = True

        # Mock torch.load to simulate loading a model
        fake_model = MagicMock(spec=torch.nn.Module)
        mock_torch_load.return_value = fake_model

        # Call the function
        model = download_model(
            ModelPaths.pfns4bo_hebo,
            cache_dir=os.environ.get("RUNNER_TEMP", "/tmp") + "/test_cache",
            # $RUNNER_TEMP is set by GitHub Actions as tmp, /tmp does not work there
        )

        # Assertions for cache hit
        # mock_exists is called once here and once through os.makedirs
        # which checks if directory exists
        self.assertEqual(mock_exists.call_count, 2)
        mock_torch_load.assert_called_once()
        self.assertEqual(model, fake_model)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass

import torch
from botorch.acquisition import AcquisitionFunction, FixedFeatureAcquisitionFunction
from botorch.optim.parameter_constraints import (
    _generate_unfixed_lin_constraints,
    _generate_unfixed_nonlin_constraints,
)
from torch import Tensor


def _flip_sub_unique(x: Tensor, k: int) -> Tensor:
    """Get the first k unique elements of a single-dimensional tensor, traversing the
    tensor from the back.

    Args:
        x: A single-dimensional tensor
        k: the number of elements to return

    Returns:
        A tensor with min(k, |x|) elements.

    Example:
        >>> x = torch.tensor([1, 6, 4, 3, 6, 3])
        >>> y = _flip_sub_unique(x, 3)  # tensor([3, 6, 4])
        >>> y = _flip_sub_unique(x, 4)  # tensor([3, 6, 4, 1])
        >>> y = _flip_sub_unique(x, 10)  # tensor([3, 6, 4, 1])

    NOTE: This should really be done in C++ to speed up the loop. Also, we would like
    to make this work for arbitrary batch shapes, I'm sure this can be sped up.
    """
    n = len(x)
    i = 0
    out = set()
    idcs = torch.empty(k, dtype=torch.long)
    for j, xi in enumerate(x.flip(0).tolist()):
        if xi not in out:
            out.add(xi)
            idcs[i] = n - 1 - j
            i += 1
        if len(out) >= k:
            break
    return x[idcs[: len(out)]]


@dataclass(frozen=True, repr=False, eq=False)
class _NoFixedFeatures:
    """
    Dataclass to store the objects after removing fixed features.
    Objects here refer to the acquisition function, initial conditions,
    bounds and parameter constraints.
    """

    acquisition_function: FixedFeatureAcquisitionFunction
    initial_conditions: Tensor
    lower_bounds: float | Tensor | None
    upper_bounds: float | Tensor | None
    inequality_constraints: list[tuple[Tensor, Tensor, float]] | None
    equality_constraints: list[tuple[Tensor, Tensor, float]] | None
    nonlinear_inequality_constraints: list[Callable[[Tensor], Tensor]] | None


def _remove_fixed_features_from_optimization(
    fixed_features: dict[int, float | None],
    acquisition_function: AcquisitionFunction,
    initial_conditions: Tensor | None,
    d: int,
    lower_bounds: float | Tensor | None,
    upper_bounds: float | Tensor | None,
    inequality_constraints: list[tuple[Tensor, Tensor, float]] | None,
    equality_constraints: list[tuple[Tensor, Tensor, float]] | None,
    nonlinear_inequality_constraints: list[Callable[[Tensor], Tensor]] | None,
) -> _NoFixedFeatures:
    """
    Given a set of non-empty fixed features, this function effectively reduces the
    dimensionality of the domain that the acquisition function is being optimized
    over by removing the set of fixed features. Consequently, this function returns a
    new `FixedFeatureAcquisitionFunction`, new constraints, and bounds defined over
    unfixed features.

    Args:
        fixed_features: This is a dictionary of feature indices to values, where
            all generated candidates will have features fixed to these values.
            Assumes values to be compatible with lower_bounds and upper_bounds!
        acquisition_function: Acquisition function over the original domain being
            maximized.
        initial_conditions: Starting points for optimization w.r.t. the complete domain.
        d: Dimensionality of the original domain.
        lower_bounds: Minimum values for each column of initial_conditions.
        upper_bounds: Minimum values for each column of initial_conditions.
        inequality constraints: A list of tuples (indices, coefficients, rhs),
            with each tuple encoding an inequality constraint of the form
            `sum_i (X[indices[i]] * coefficients[i]) >= rhs`.
        equality constraints: A list of tuples (indices, coefficients, rhs),
            with each tuple encoding an inequality constraint of the form
            `sum_i (X[indices[i]] * coefficients[i]) = rhs`.
        nonlinear_inequality_constraints: A list of callables with that represent
            non-linear inequality constraints of the form `callable(x) >= 0`. Each
            callable is expected to take a `(num_restarts) x q x d`-dim tensor as
            an input and return a `(num_restarts) x q`-dim tensor with the
            constraint values.

    Returns:
        _NoFixedFeatures dataclass object.
    """
    if not fixed_features:
        return _NoFixedFeatures(
            acquisition_function=acquisition_function,
            initial_conditions=initial_conditions,
            lower_bounds=lower_bounds,
            upper_bounds=upper_bounds,
            inequality_constraints=inequality_constraints,
            equality_constraints=equality_constraints,
            nonlinear_inequality_constraints=nonlinear_inequality_constraints,
        )
    # sort the keys for consistency
    sorted_keys = sorted(fixed_features)
    sorted_values = []
    for key in sorted_keys:
        val = fixed_features[key]
        sorted_values.append(val)

    acquisition_function = FixedFeatureAcquisitionFunction(
        acq_function=acquisition_function,
        d=d,
        columns=sorted_keys,
        values=sorted_values,
    )

    # extract initial_conditions, bounds at unfixed indices
    unfixed_indices = sorted(set(range(d)) - set(sorted_keys))
    if initial_conditions is not None:
        initial_conditions = initial_conditions[..., unfixed_indices]
    if isinstance(lower_bounds, Tensor):
        lower_bounds = lower_bounds[..., unfixed_indices]
    if isinstance(upper_bounds, Tensor):
        upper_bounds = upper_bounds[..., unfixed_indices]

    inequality_constraints = _generate_unfixed_lin_constraints(
        constraints=inequality_constraints,
        fixed_features=fixed_features,
        dimension=d,
        eq=False,
    )
    equality_constraints = _generate_unfixed_lin_constraints(
        constraints=equality_constraints,
        fixed_features=fixed_features,
        dimension=d,
        eq=True,
    )
    nonlinear_inequality_constraints = _generate_unfixed_nonlin_constraints(
        constraints=nonlinear_inequality_constraints,
        fixed_features=fixed_features,
        dimension=d,
    )
    return _NoFixedFeatures(
        acquisition_function=acquisition_function,
        initial_conditions=initial_conditions,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        inequality_constraints=inequality_constraints,
        equality_constraints=equality_constraints,
        nonlinear_inequality_constraints=nonlinear_inequality_constraints,
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Helpers for multitask modeling.
"""

from __future__ import annotations

import torch
from gpytorch.distributions import MultitaskMultivariateNormal
from gpytorch.distributions.multivariate_normal import MultivariateNormal
from linear_operator import to_linear_operator


def separate_mtmvn(mvn: MultitaskMultivariateNormal) -> list[MultivariateNormal]:
    """
    Separate a MTMVN into a list of MVNs, where covariance across data within each task
    are preserved, while covariance across task are dropped.
    """
    # T150340766 Upstream as a class method on gpytorch MultitaskMultivariateNormal.
    full_covar = mvn.lazy_covariance_matrix
    num_data, num_tasks = mvn.mean.shape[-2:]

    mvns = []
    for c in range(num_tasks):
        # Compute indices for task c's data points
        if mvn._interleaved:
            # For interleaved: task c data points are at positions
            # c, c+num_tasks, c+2*num_tasks, ...
            task_indices = torch.arange(
                c, num_data * num_tasks, num_tasks, device=full_covar.device
            )
        else:
            # For non-interleaved: task c data points are at positions
            # c*num_data to (c+1)*num_data
            task_indices = torch.arange(
                c * num_data, (c + 1) * num_data, device=full_covar.device
            )

        # Extract covariance submatrix for task c
        task_covar = full_covar[..., task_indices, :]
        task_covar = task_covar[..., :, task_indices]

        mvns.append(
            MultivariateNormal(mvn.mean[..., c], to_linear_operator(task_covar))
        )
    return mvns



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Helper utilities for constructing scalarizations.

References

.. [Knowles2005]
    J. Knowles, "ParEGO: a hybrid algorithm with on-line landscape approximation
    for expensive multiobjective optimization problems," in IEEE Transactions
    on Evolutionary Computation, vol. 10, no. 1, pp. 50-66, Feb. 2006.
"""

from __future__ import annotations

from collections.abc import Callable

import torch
from botorch.exceptions.errors import BotorchTensorDimensionError, UnsupportedError
from botorch.utils.transforms import normalize
from torch import Tensor


def get_chebyshev_scalarization(
    weights: Tensor, Y: Tensor, alpha: float = 0.05
) -> Callable[[Tensor, Tensor | None], Tensor]:
    r"""Construct an augmented Chebyshev scalarization.

    The augmented Chebyshev scalarization is given by
        g(y) = max_i(w_i * y_i) + alpha * sum_i(w_i * y_i)

    where the goal is to minimize g(y) in the setting where all objectives y_i are
    to be minimized. Since the default in BoTorch is to maximize all objectives,
    this method constructs a Chebyshev scalarization where the inputs are first
    multiplied by -1, so that all objectives are to be minimized. Then, it computes
    g(y) (which should be minimized), and returns -g(y), which should be maximized.

    Minimizing an objective is supported by passing a negative
    weight for that objective. To make all w * y's have the same sign
    such that they are comparable when computing max(w * y), outcomes of minimization
    objectives are shifted from [0,1] to [-1,0].

    See [Knowles2005]_ for details.

    This scalarization can be used with qExpectedImprovement to implement q-ParEGO
    as proposed in [Daulton2020qehvi]_.

    Args:
        weights: A `m`-dim tensor of weights.
            Positive for maximization and negative for minimization.
        Y: A `n x m`-dim tensor of observed outcomes, which are used for
            scaling the outcomes to [0,1] or [-1,0]. If `n=0`, then outcomes
            are left unnormalized.
        alpha: Parameter governing the influence of the weighted sum term. The
            default value comes from [Knowles2005]_.

    Returns:
        Transform function using the objective weights.

    Example:
        >>> weights = torch.tensor([0.75, -0.25])
        >>> transform = get_aug_chebyshev_scalarization(weights, Y)
    """
    # the chebyshev_obj assumes all objectives should be minimized, so
    # multiply Y by -1
    Y = -Y
    if weights.shape != Y.shape[-1:]:
        raise BotorchTensorDimensionError(
            "weights must be an `m`-dim tensor where Y is `... x m`."
            f"Got shapes {weights.shape} and {Y.shape}."
        )
    elif Y.ndim > 2:
        raise NotImplementedError("Batched Y is not currently supported.")

    def chebyshev_obj(Y: Tensor, X: Tensor | None = None) -> Tensor:
        product = weights * Y
        return product.max(dim=-1).values + alpha * product.sum(dim=-1)

    # A boolean mask indicating if minimizing an objective
    minimize = weights < 0
    if Y.shape[-2] == 0:
        if minimize.any():
            raise UnsupportedError(
                "negative weights (for minimization) are only supported if "
                "Y is provided."
            )
        # If there are no observations, we do not need to normalize the objectives

        def obj(Y: Tensor, X: Tensor | None = None) -> Tensor:
            # multiply the scalarization by -1, so that the scalarization should
            # be maximized
            return -chebyshev_obj(Y=-Y)

        return obj
    # Set the bounds to be [min(Y_m), max(Y_m)], for each objective m.
    Y_bounds = torch.stack([Y.min(dim=-2).values, Y.max(dim=-2).values])

    def obj(Y: Tensor, X: Tensor | None = None) -> Tensor:
        # scale to [0,1]
        Y_normalized = normalize(-Y, bounds=Y_bounds)
        # If minimizing an objective, convert Y_normalized values to [-1,0],
        # such that min(w*y) makes sense, we want all w*y's to be positive
        Y_normalized[..., minimize] = Y_normalized[..., minimize] - 1
        # multiply the scalarization by -1, so that the scalarization should
        # be maximized
        return -chebyshev_obj(Y=Y_normalized)

    return obj



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""Utilities for interfacing Numpy and Torch."""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import torch
from torch import Tensor


torch_to_numpy_dtype_dict = {
    torch.bool: bool,
    torch.uint8: np.uint8,
    torch.int8: np.int8,
    torch.int16: np.int16,
    torch.int32: np.int32,
    torch.int64: np.int64,
    torch.float16: np.float16,
    torch.float32: np.float32,
    torch.float64: np.float64,
    torch.complex64: np.complex64,
    torch.complex128: np.complex128,
}


def as_ndarray(
    values: Tensor, dtype: np.dtype | None = None, inplace: bool = True
) -> npt.NDArray:
    r"""Helper for going from torch.Tensor to numpy.ndarray.

    Args:
        values: Tensor to be converted to ndarray.
        dtype: Optional numpy.dtype for the converted tensor.
        inplace: Boolean indicating whether memory should be shared if possible.

    Returns:
        An ndarray with the same data as `values`.
    """
    with torch.no_grad():
        out = values.cpu()  # maybe transfer to cpu

        # Determine whether or not to `clone`
        if (
            # cond 1: are we not in `inplace` mode?
            not inplace
            # cond 2: did we already copy when calling `cpu` above?
            and out.device == values.device
            # cond 3: will we copy when calling `astype` below?
            and (dtype is None or out.dtype == torch_to_numpy_dtype_dict[dtype])
        ):
            out = out.clone()

        # Convert to ndarray and maybe cast to `dtype`
        out = out.numpy()
        return out.astype(dtype, copy=False)


def get_bounds_as_ndarray(
    parameters: dict[str, Tensor],
    bounds: dict[str, tuple[float | Tensor | None, float | Tensor | None]],
) -> npt.NDArray | None:
    r"""Helper method for converting bounds into an ndarray.

    Args:
        parameters: A dictionary of parameters.
        bounds: A dictionary of (optional) lower and upper bounds.

    Returns:
        An ndarray of bounds.
    """
    inf = float("inf")
    full_size = sum(param.numel() for param in parameters.values())
    out = np.full((full_size, 2), (-inf, inf))
    index = 0
    for name, param in parameters.items():
        size = param.numel()
        if name in bounds:
            lower, upper = bounds[name]
            lower = -inf if lower is None else lower
            upper = inf if upper is None else upper
            if isinstance(lower, Tensor):
                lower = lower.cpu()
            if isinstance(upper, Tensor):
                upper = upper.cpu()
            out[index : index + size, 0] = lower
            out[index : index + size, 1] = upper
        index = index + size
    # If all bounds are +/- inf, return None.
    if np.isinf(out).all():
        out = None
    return out



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Abstract class for acquisition functions leveraging a cached Cholesky
decomposition of the posterior covariance over f(X_baseline).
"""

from __future__ import annotations

import warnings

import torch
from botorch.acquisition.acquisition import MCSamplerMixin
from botorch.exceptions.warnings import BotorchWarning
from botorch.models.gpytorch import GPyTorchModel
from botorch.models.higher_order_gp import HigherOrderGP
from botorch.models.model import Model
from botorch.models.model_list_gp_regression import ModelListGP
from botorch.models.multitask import KroneckerMultiTaskGP, MultiTaskGP
from botorch.posteriors.gpytorch import GPyTorchPosterior
from botorch.posteriors.posterior import Posterior
from botorch.sampling.base import MCSampler
from botorch.utils.low_rank import extract_batch_covar, sample_cached_cholesky
from gpytorch.distributions.multitask_multivariate_normal import (
    MultitaskMultivariateNormal,
)
from linear_operator.utils.errors import NanError, NotPSDError
from torch import Tensor


def supports_cache_root(model: Model) -> bool:
    r"""Checks if a model supports the cache_root functionality.
    The two criteria are that the model is not multi-task and the model
    produces a GPyTorchPosterior.
    """
    if isinstance(model, ModelListGP):
        return all(supports_cache_root(m) for m in model.models)
    # Multi task models and non-GPyTorch models are not supported.
    if isinstance(
        model, (MultiTaskGP, KroneckerMultiTaskGP, HigherOrderGP)
    ) or not isinstance(model, GPyTorchModel):
        return False
    # Models that return a TransformedPosterior are not supported.
    if hasattr(model, "outcome_transform") and (not model.outcome_transform._is_linear):
        return False
    return True


def _get_cache_root_not_supported_message(model_cls: type) -> str:
    msg = (
        "`cache_root` is only supported for GPyTorchModels that "
        "are not MultiTask models and don't produce a "
        f"TransformedPosterior. Got a model of type {model_cls}. Setting "
        "`cache_root = False`."
    )
    return msg


class CachedCholeskyMCSamplerMixin(MCSamplerMixin):
    r"""Abstract Mixin class for acquisition functions using a cached Cholesky.

    Specifically, this is for acquisition functions that require sampling from
    the posterior P(f(X_baseline, X) | D). The Cholesky of the posterior
    covariance over f(X_baseline) is cached.
    """

    def __init__(
        self,
        model: Model,
        cache_root: bool = False,
        sampler: MCSampler | None = None,
    ) -> None:
        r"""Set class attributes and perform compatibility checks.

        Args:
            model: A model.
            cache_root: A boolean indicating whether to cache the Cholesky.
                This might be overridden in the model is not compatible.
            sampler: An optional MCSampler object.
        """
        MCSamplerMixin.__init__(self, sampler=sampler)
        if cache_root and not supports_cache_root(model):
            warnings.warn(
                _get_cache_root_not_supported_message(type(model)),
                RuntimeWarning,
                stacklevel=3,
            )
            cache_root = False
        self._cache_root = cache_root

    def _compute_root_decomposition(
        self,
        posterior: Posterior,
    ) -> Tensor:
        r"""Cache Cholesky of the posterior covariance over f(X_baseline).

        Because `LinearOperator.root_decomposition` is decorated with LinearOperator's
        @cached decorator, this function is doing a lot implicitly:

        1) Check if a root decomposition has already been cached to `lazy_covar`.
          Note that it will not have been if `posterior.mvn` is a
          `MultitaskMultivariateNormal`, since we construct `lazy_covar` in that
          case.
        2) If the root decomposition has not been found in the cache, compute it.
        3) Write it to the cache of `lazy_covar`. Note that this will become
          inaccessible if `posterior.mvn` is a `MultitaskMultivariateNormal`,
          since in that case `lazy_covar`'s scope is only this function.

        Args:
            posterior: The posterior over f(X_baseline).
        """
        if isinstance(posterior.distribution, MultitaskMultivariateNormal):
            lazy_covar = extract_batch_covar(posterior.distribution)
        else:
            lazy_covar = posterior.distribution.lazy_covariance_matrix
        lazy_covar_root = lazy_covar.root_decomposition()
        return lazy_covar_root.root.to_dense()

    def _get_f_X_samples(self, posterior: GPyTorchPosterior, q_in: int) -> Tensor:
        r"""Get posterior samples at the `q_in` new points from the joint posterior.

        Args:
            posterior: The joint posterior is over (X_baseline, X).
            q_in: The number of new points in the posterior. See `_set_sampler` for
                more information.

        Returns:
            A `sample_shape x batch_shape x q x m`-dim tensor of posterior
                samples at the new points.
        """
        # Technically we should make sure that we add a consistent nugget to the
        # cached covariance (and box decompositions) and the new block.
        # But recomputing box decompositions every time the jitter changes would
        # be quite slow.
        if self._cache_root and hasattr(self, "_baseline_L"):
            try:
                return sample_cached_cholesky(
                    posterior=posterior,
                    baseline_L=self._baseline_L,
                    q=q_in,
                    base_samples=self.sampler.base_samples,
                    sample_shape=self.sampler.sample_shape,
                )
            except (NanError, NotPSDError):
                warnings.warn(
                    "Low-rank cholesky updates failed due NaNs or due to an "
                    "ill-conditioned covariance matrix. "
                    "Falling back to standard sampling.",
                    BotorchWarning,
                    stacklevel=3,
                )

        # TODO: improve efficiency for multi-task models
        samples = self.get_posterior_samples(posterior)
        if isinstance(self.model, HigherOrderGP):
            # Select the correct q-batch dimension for HOGP.
            q_dim = -self.model._num_dimensions
            q_idcs = (
                torch.arange(-q_in, 0, device=samples.device) + samples.shape[q_dim]
            )
            return samples.index_select(q_dim, q_idcs)
        else:
            return samples[..., -q_in:, :]

    def _set_sampler(
        self,
        q_in: int,
        posterior: Posterior,
    ) -> None:
        r"""Update the sampler to use the original base samples for X_baseline.

        Args:
            q_in: The effective input batch size. This is typically equal to the
                q-batch size of `X`. However, if using a one-to-many input transform,
                e.g., `InputPerturbation` with `n_w` perturbations, the posterior will
                have `n_w` points on the q-batch for each point on the q-batch of `X`.
                In which case, `q_in = q * n_w` is used.
            posterior: The posterior.
        """
        if self.q_in != q_in and self.base_sampler is not None:
            self.sampler._update_base_samples(
                posterior=posterior, base_sampler=self.base_sampler
            )
            self.q_in = q_in



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Abstract base module for all botorch posteriors.
"""

from __future__ import annotations

from typing import Callable, Optional, Union

import torch
from botorch.posteriors.posterior import Posterior
from torch import Tensor


class BoundedRiemannPosterior(Posterior):
    """
    Notes: Bounded posterior for now, will work on unbounded posteriors.
    This is also only over 1 test point, not batches.
    """

    def __init__(self, borders, probabilities):
        r"""Bounded Riemann Posterior.

        A posterior distribution represented by a piecewise constant probability density
        function over a bounded domain. The domain is divided into buckets defined by
        borders, with each bucket having an associated probability.

        Args:
            borders: A tensor of shape `(n_buckets + 1,)` defining the boundaries of
                the buckets. Must be monotonically increasing.
            probabilities: A tensor of shape `(..., n_buckets,)` defining the
                probability mass in each bucket. Must sum to 1 in the last dim.
        """

        assert torch.allclose(
            probabilities.sum(-1),
            torch.tensor(1.0, device=probabilities.device, dtype=probabilities.dtype),
            atol=1e-3,
        ), f"Probabilities must sum to 1, but sum to {probabilities.sum()}."

        self.borders = borders
        self.probabilities = probabilities
        self.cumprobs = torch.cumsum(self.probabilities, -1)

    def integrate(self, ag_integrate_fn: Callable[[Tensor, Tensor], Tensor]) -> Tensor:
        r"""Integrate over the posterior, by calculating
            $$\int_{min}^{max} ag(y) posterior(y) dy$$.

        Args:
            ag_integral_fn: A vectorized function that integrates ag from lower_bound
                to upper_bound, that is, $$\int_{lower_bound}^{upper_bound} ag(y) dy$$.

        Returns:
            Tensor: The integral of the posterior.
        """
        all_lower = self.borders[:-1]
        all_upper = self.borders[1:]
        bucket_results = ag_integrate_fn(all_lower, all_upper)
        return (bucket_results * (self.probabilities / (all_upper - all_lower))).sum(-1)

    def rsample(
        self,
        sample_shape: Optional[torch.Size] = None,
    ) -> Tensor:
        r"""Sample from the posterior (with gradients).

        Args:
            sample_shape: A `torch.Size` object specifying the sample shape. To
                draw `n` samples, set to `torch.Size([n])`. To draw `b` batches
                of `n` samples each, set to `torch.Size([b, n])`.

        Returns:
            Samples from the posterior, a tensor of shape
            `self._extended_shape(sample_shape=sample_shape)`.
        """
        sample_shape = sample_shape if sample_shape is not None else torch.Size([1])
        z = torch.rand(sample_shape)
        return self.rsample_from_base_samples(sample_shape, z)

    def rsample_from_base_samples(
        self, sample_shape: torch.Size, base_samples: Tensor
    ) -> Tensor:
        if base_samples.shape[: len(sample_shape)] != sample_shape:
            raise RuntimeError(
                "`sample_shape` disagrees with shape of `base_samples`. "
                f"Got {sample_shape=} and {base_samples.shape=}."
            )
        return self.icdf(base_samples)

    @property
    def device(self) -> torch.device:
        r"""The torch device of the distribution."""
        return self.borders.device

    @property
    def dtype(self) -> torch.dtype:
        r"""The torch dtype of the distribution."""
        return self.borders.dtype

    @property
    def mean(self):
        r"""The mean of the posterior distribution."""
        bucket_widths = self.borders[1:] - self.borders[:-1]
        bucket_means = self.borders[:-1] + bucket_widths / 2
        return (self.probabilities @ bucket_means).unsqueeze(-1)

    @property
    def mean_of_square(self) -> torch.Tensor:
        """Computes E[x^2]."""
        left_borders = self.borders[:-1]
        right_borders = self.borders[1:]
        bucket_mean_of_square = (
            left_borders.square()
            + right_borders.square()
            + left_borders * right_borders
        ) / 3.0
        return (self.probabilities @ bucket_mean_of_square).unsqueeze(-1)

    @property
    def variance(self) -> torch.Tensor:
        """Computes the variance via Var[x] = E[x^2] - E[x]^2."""
        return self.mean_of_square - self.mean.square()

    def confidence_region(
        self, confidence_level: float = 0.95
    ) -> tuple[Tensor, Tensor]:
        r"""
        Compute the lower and upper bounds of the confidence region.
        Args:
            confidence_level: The probability between the bounds
                of the confidence interval/region.
                Use .954 for 2 sigma of a normal distribution.
        """
        side_probs = (1.0 - confidence_level) / 2
        return self.icdf(side_probs), self.icdf(1.0 - side_probs)

    def icdf(self, value: Union[Tensor, float]) -> Tensor:
        r"""Inverse cdf (with gradients).
        Use value to get the index of the bucket that contains the value
        and then interpolate between the left and right borders of the bucket

        Args:
            value: The value at which to evaluate the inverse CDF.

        Returns:
            The inverse CDF of the posterior at the given value(s).
            The shape of the return is the shape of value, with the batch
            shape of the probs (all dims up to the final dim) appended
            with a final trailing dimension of 1, for the dim of the dist.
        """

        # final shape is (batch_shape, -1)
        value = torch.as_tensor(
            value, device=self.borders.device, dtype=self.borders.dtype
        )
        value_shape = value.shape
        # shape of cumprobs is (batch_shape, n_buckets)
        value = value.broadcast_to(size=(*self.cumprobs.shape[:-1], *value_shape))
        value = value.reshape(*self.cumprobs.shape[:-1], -1)

        # get first index where cumprobs > value
        index = torch.searchsorted(self.cumprobs, value)

        left_border = self.borders[index]
        right_border = self.borders[index + 1]

        bucket_width = right_border - left_border
        right_cum_probs = torch.gather(self.cumprobs, -1, index)
        prob_width = torch.gather(self.probabilities, -1, index)

        bucket_proportion_remaining = (right_cum_probs - value) / prob_width
        result = left_border + (1 - bucket_proportion_remaining) * bucket_width

        # reshape to (value_shape, batch_shape, 1)
        result = result.transpose(0, -1)
        return result.reshape(*value_shape, *self.cumprobs.shape[:-1], 1)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""Core abstractions and generic optimizers."""

from __future__ import annotations

import re
from collections.abc import Callable, Sequence
from dataclasses import dataclass, replace
from enum import auto, Enum
from itertools import count
from sys import maxsize
from time import monotonic
from typing import Any

import numpy.typing as npt

from botorch.optim.closures import NdarrayOptimizationClosure
from botorch.optim.stopping import StoppingCriterion
from botorch.optim.utils.numpy_utils import get_bounds_as_ndarray
from botorch.optim.utils.timeout import minimize_with_timeout
from numpy import asarray, float64 as np_float64
from torch import Tensor
from torch.optim.adam import Adam
from torch.optim.optimizer import Optimizer

try:
    from torch.optim.lr_scheduler import LRScheduler
except ImportError:  # pragma: no cover
    from torch.optim.lr_scheduler import _LRScheduler as LRScheduler  # pragma: no cover


_LBFGSB_MAXITER_MAXFUN_REGEX = re.compile(  # regex for maxiter and maxfun messages
    # Note that the messages changed with scipy 1.15, hence the different matching here.
    "TOTAL NO. (of|OF) "
    + "(ITERATIONS REACHED LIMIT|(f AND g|F,G) EVALUATIONS EXCEEDS LIMIT)"
)


class OptimizationStatus(int, Enum):
    RUNNING = auto()  # incomplete
    SUCCESS = auto()  # optimizer converged
    FAILURE = auto()  # terminated abnormally
    STOPPED = auto()  # stopped due to user provided criterion


@dataclass
class OptimizationResult:
    step: int
    fval: float | int
    status: OptimizationStatus
    runtime: float | None = None
    message: str | None = None


def scipy_minimize(
    closure: (
        Callable[[], tuple[Tensor, Sequence[Tensor | None]]]
        | NdarrayOptimizationClosure
    ),
    parameters: dict[str, Tensor],
    bounds: dict[str, tuple[float | None, float | None]] | None = None,
    callback: Callable[[dict[str, Tensor], OptimizationResult], None] | None = None,
    x0: npt.NDArray | None = None,
    method: str = "L-BFGS-B",
    options: dict[str, Any] | None = None,
    timeout_sec: float | None = None,
) -> OptimizationResult:
    r"""Generic scipy.optimize.minimize-based optimization routine.

    Args:
        closure: Callable that returns a tensor and an iterable of gradient tensors or
            NdarrayOptimizationClosure instance.
        parameters: A dictionary of tensors to be optimized.
        bounds: A dictionary mapping parameter names to lower and upper bounds.
        callback: A callable taking `parameters` and an OptimizationResult as arguments.
        x0: An optional initialization vector passed to scipy.optimize.minimize.
        method: Solver type, passed along to scipy.optimize.minimize.
        options: Dictionary of solver options, passed along to scipy.optimize.minimize.
        timeout_sec: Timeout in seconds to wait before aborting the optimization loop
            if not converged (will return the best found solution thus far).

    Returns:
        An OptimizationResult summarizing the final state of the run.
    """
    start_time = monotonic()
    wrapped_closure = (
        closure
        if isinstance(closure, NdarrayOptimizationClosure)
        else NdarrayOptimizationClosure(closure, parameters)
    )
    if bounds is None:
        bounds_np = None
    else:
        bounds_np = get_bounds_as_ndarray(parameters, bounds)

    if callback is None:
        wrapped_callback = None
    else:
        call_counter = count(1)  # callbacks are typically made at the end of each iter

        def wrapped_callback(x: npt.NDArray):
            result = OptimizationResult(
                step=next(call_counter),
                fval=float(wrapped_closure(x)[0]),
                status=OptimizationStatus.RUNNING,
                runtime=monotonic() - start_time,
            )
            return callback(parameters, result)  # pyre-ignore [29]

    raw = minimize_with_timeout(
        wrapped_closure,
        wrapped_closure.state if x0 is None else x0.astype(np_float64, copy=False),
        jac=True,
        bounds=bounds_np,
        method=method,
        options=options,
        callback=wrapped_callback,
        timeout_sec=timeout_sec,
    )

    # Post-processing and outcome handling
    wrapped_closure.state = asarray(raw.x)  # set parameter state to optimal values
    msg = raw.message if isinstance(raw.message, str) else raw.message.decode("ascii")
    if raw.success:
        status = OptimizationStatus.SUCCESS
    else:
        status = (  # Check whether we stopped due to reaching maxfun or maxiter
            OptimizationStatus.STOPPED
            if _LBFGSB_MAXITER_MAXFUN_REGEX.search(msg)
            or "Optimization timed out after" in msg
            else OptimizationStatus.FAILURE
        )

    return OptimizationResult(
        fval=raw.fun,
        step=raw.nit,
        status=status,
        message=msg,
        runtime=monotonic() - start_time,
    )


def torch_minimize(
    closure: Callable[[], tuple[Tensor, Sequence[Tensor | None]]],
    parameters: dict[str, Tensor],
    bounds: dict[str, tuple[float | None, float | None]] | None = None,
    callback: Callable[[dict[str, Tensor], OptimizationResult], None] | None = None,
    optimizer: Optimizer | Callable[[list[Tensor]], Optimizer] = Adam,
    scheduler: LRScheduler | Callable[[Optimizer], LRScheduler] | None = None,
    step_limit: int | None = None,
    timeout_sec: float | None = None,
    stopping_criterion: StoppingCriterion | None = None,
) -> OptimizationResult:
    r"""Generic torch.optim-based optimization routine.

    Args:
        closure: Callable that returns a tensor and an iterable of gradient tensors.
            Responsible for setting relevant parameters' `grad` attributes.
        parameters: A dictionary of tensors to be optimized.
        bounds: An optional dictionary of bounds for elements of `parameters`.
        callback: A callable taking `parameters` and an OptimizationResult as arguments.
        optimizer: A `torch.optim.Optimizer` instance or a factory that takes
            a list of parameters and returns an `Optimizer` instance.
        scheduler: A `torch.optim.lr_scheduler._LRScheduler` instance or a factory
            that takes a `Optimizer` instance and returns a `_LRSchedule` instance.
        step_limit: Integer specifying a maximum number of optimization steps.
            One of `step_limit`, `stopping_criterion`, or `timeout_sec` must be passed.
        timeout_sec: Timeout in seconds before terminating the optimization loop.
            One of `step_limit`, `stopping_criterion`, or `timeout_sec` must be passed.
        stopping_criterion: A StoppingCriterion for the optimization loop.

    Returns:
        An OptimizationResult summarizing the final state of the run.
    """
    result: OptimizationResult
    start_time = monotonic()

    if step_limit is None:
        if stopping_criterion is None and timeout_sec is None:
            raise RuntimeError("No termination conditions were given.")
        step_limit = maxsize

    if not isinstance(optimizer, Optimizer):
        optimizer = optimizer(list(parameters.values()))

    if not (scheduler is None or isinstance(scheduler, LRScheduler)):
        scheduler = scheduler(optimizer)

    if stopping_criterion is not None:
        # Reset stopping criterion to ensure clean state for new optimization run
        stopping_criterion.reset()

    _bounds = (
        {}
        if bounds is None
        else {name: limits for name, limits in bounds.items() if name in parameters}
    )
    for step in range(1, step_limit + 1):
        fval = closure()[0].detach()
        runtime = monotonic() - start_time
        result = OptimizationResult(
            step=step,
            fval=fval.cpu().item(),
            status=OptimizationStatus.RUNNING,
            runtime=runtime,
        )

        # TODO: Update stopping_criterion API to return a message.
        if stopping_criterion and stopping_criterion(fval):
            result.status = OptimizationStatus.STOPPED
            result.message = "`torch_minimize` stopped due to `stopping_criterion`."

        if timeout_sec is not None and runtime >= timeout_sec:
            result.status = OptimizationStatus.STOPPED
            result.message = (
                f"`torch_minimize` stopped due to timeout after {runtime} seconds."
            )

        if callback:
            callback(parameters, result)

        if result.status != OptimizationStatus.RUNNING:
            break

        optimizer.step()
        for name, (lower, upper) in _bounds.items():
            parameters[name].data = parameters[name].clamp(min=lower, max=upper)

        if scheduler:
            scheduler.step()

    if result.status != OptimizationStatus.RUNNING:
        return replace(result, runtime=monotonic() - start_time)

    # Account for final parameter update when stopping due to step_limit
    return OptimizationResult(
        step=step,
        fval=closure()[0].detach().cpu().item(),
        status=OptimizationStatus.STOPPED,
        runtime=monotonic() - start_time,
        message=f"`torch_minimize` stopped after reaching step_limit={step_limit}.",
    )
